# -----------------------------------------------------------------------------
# Copyright (C) 2013-2022 Michael I. Love, Constantin Ahlmann-Eltze
# Copyright (C) 2023 Maximilien Colange

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
# -----------------------------------------------------------------------------

# This file is based on the file 'R/results.R' of the Bioconductor DESeq2
# package (version 3.16).


import numpy as np
import pandas as pd
import patsy
import scipy.stats
from statsmodels.stats.multitest import multipletests
from statsmodels.nonparametric.smoothers_lowess import lowess

from ..utils import pnorm, pt
from .. import __version__
from .deseq2_cpp import fitBeta
from .misc import buildDataFrameWithNACols, getFactorName


def p_adjust(*args, **kwargs):
    """
    Test results and p-value correction for multiple tests

    This is a wrapper around :func:`statsmodels.stats.multitest.multipletests`
    """
    return multipletests(*args, **kwargs)[1]


class DESeqResults(pd.DataFrame):
    """
    a class to store a results table

    This class would not typically constructed directly by "end users". This
    simple class extends :class:`pandas.DataFrame`. It is used by
    :meth:`.DESeqDataSet.results` to wrap up the results table.

    Attributes
    ----------
    priorInfo : array-like
        a list giving information on the log fold change prior
    """

    from .plot import plotMA_res as plotMA

    _metadata = [
        "priorInfo",
        "filterThreshold",
        "filterTheta",
        "filterNumRej",
        "lo_fit",
        "alpha",
    ]

    @property
    def _constructor(self):
        return DESeqResults

    @property
    def _constructor_sliced(self):
        return pd.Series

    def __init__(self, df, priorInfo=None, *args, **kwargs):
        super().__init__(df, *args, **kwargs)
        if priorInfo is not None:
            self.priorInfo = priorInfo

    def summary(self, alpha=None):
        """
        Print a summary of the results from a DESeq analysis.

        Arguments
        ---------
        alpha : float, optional
            the adjusted p-value cutoff. If not set, this defaults to the
            :code:`alpha` argument which was used in :meth:`results` to set the
            target FDR for independent filtering, or if independent filtering
            was not performed, to 0.1.

        Returns
        -------
        str
            a summary (multi-line) string
        """
        sval = "svalue" in self.columns
        if sval:
            test_col = "svalue"
            test_col_name = "s-value"
        else:
            test_col = "padj"
            test_col_name = "adjusted p-value"

        if alpha is None:
            if sval:
                alpha = 0.005
            elif self.alpha is None:
                alpha = 0.1
            else:
                alpha = self.alpha

        if self.lfcThreshold is not None:
            T = self.lfcThreshold
            pT = f"{T:.2f} (up)    "
            mT = f"{-T:.2f} (down) "
        else:
            T = 0
        if T == 0:
            pT = "0 (up)       "
            mT = "0 (down)     "

        res = "\n"
        notallzero = (self.baseMean > 0).sum()
        up = np.nansum((self[test_col] < alpha) & (self["log2FoldChange"] > T))
        down = np.nansum((self[test_col] < alpha) & (self["log2FoldChange"] < -T))
        if not sval:
            filt = (~(self["pvalue"].isna()) & self["padj"].isna()).sum()
            outlier = ((self["baseMean"] > 0) & self["pvalue"].isna()).sum()
            if self.filterThreshold is None:
                ft = 0
            else:
                ft = round(self.filterThreshold)

        ihw = (not sval) and "ihwResult" in self.columns

        res += f"out of {notallzero} with nonzero total read count\n"
        res += f"{test_col_name} < {alpha}\n"
        res += f"LFC > {pT}: {up}, {up/notallzero*100:.2f}%\n"
        res += f"LFC < {mT}: {down}, {down/notallzero*100:.2f}%\n"
        if not sval:
            res += f"outliers [1]       : {outlier}, {outlier/notallzero*100:.2f}%\n"
        if not sval and not ihw:
            res += f"low counts [2]     : {filt}, {filt/notallzero*100:.2f}%\n"
        if not sval and not ihw:
            res += f"(mean count < {ft})\n"
        if not sval:
            res += "[1] see 'cooksCutoff' argument of results()\n"
        if not sval and not ihw:
            res += "[2] see 'independentFiltering' argument of results()\n"
        if ihw:
            res += "see res['ihwResult'] on hypothesis weighting\n"

        return res


def results_dds(
    obj,
    contrast=None,
    name=None,
    lfcThreshold=0,
    altHypothesis="greaterAbs",
    listValues=(1, -1),
    cooksCutoff=None,
    independentFiltering=True,
    alpha=0.1,
    filter=None,
    theta=None,
    pAdjustMethod="fdr_bh",
    filterRun=None,
    saveCols=None,
    test=None,
    addMLE=False,
    tidy=False,
    parallel=False,
    minmu=0.5,
):
    """
    Extract results from a :func:`.DESeq` analysis

    This function extracts a result table from a DESeq analysis giving base
    means across samples, log2 fold changes, standard errors, test statistics,
    p-values and adjusted p-values.

    The results table when printed will provide the information about the
    comparison, *e.g.* "log2 fold change (MAP): condition treated vs.
    untreated", meaning that the estimates are of lof2(treated / untreated), as
    would be returned using :code:`contrast=c("condition", "treated",
    "untreated")`. Multiple results can be returned for analyses beyond a
    simple two group comparison, so this function takes arguments
    :code:`contrast` and :code:`name` to help the user pick out the comparisons
    of interest for printing a results table. The use of the :code:`contrast`
    argument is recommended for exact specification of the levels which should
    be compared and their order.

    If :code:`results` is run without specifying :code:`contrast` or
    :code:`name`, it will return the comparison of the last level of the last
    variable in the design formula over the first level of this variable. For
    example, for a simple two-group comparison, this would return the log2 fold
    changes of the second group over the first group (the reference level).

    The argument :code:`contrast` can be used to generate results tables for
    any comparison of interest, for example, the log2 fold change between two
    levels of a factor, and its usage is described below. It can also
    accomodate more complicated numeric comparisons. Note that :code:`contrast`
    will set to 0 the estimated LFC in a comparison of two groups, where all
    the counts in the two groups are equal to 0 (while other groups have
    positive counts), while :code:`name` will not automatically set these LFC
    to 0.  The test statistic used for a contrast is: :math:`c^t \\beta /
    \sqrt{c^t \Sigma c}`.

    The argument :code:`name` can be used to generate results tables for
    individual effects, which must be individual elements of
    :code:`obj.resultsNames()`.  These individual effects could represent
    continuous covariates, effects for individual levels, or individual
    interaction effects.

    Information on the comparison which was used to build the results table,
    and the statistical test which was used for p-values (Wald test or
    likelihood ratio test) is stored within the object returned by
    :meth:`results()`. This information is stored in the columns of the results
    table (see :class:`.DESeqResults`).

    .. rubric:: On p-values

    By default, independent filtering is performed to select a set of genes for
    multiples test correction which maximizes the number of adjusted p-values
    less than a given critical value :code:`alpha` (by default 0.1). See the
    reference in this page for details on independent filtering. The filter
    used for maximizing the number of rejections is the mean of normalized
    counts for all samples in the dataset. Several arguments from
    :func:`.filtered_p` of the genefilter package (used within :meth:`results`)
    are provided here to control the independent filtering behavior. Note that
    the code of :func:`.filtered_p` is copied into this package to avoid extra
    dependencies).

    The threshold that is chosen is the lowest quantile of the filter for which
    the number of rejections is close to the peack of a curve fit to the number
    of rejections over the filter quantiles. "Close to" is defined as within 1
    residuel standard deviation. The adjusted p-values for the genes which do
    not pass the filter threshold are set to :code:`nan`.

    By default, :meth:`results` assigns a p-value of :code:`nan` to genes
    containing count outliers, as identified using Cook's distance. See the
    :code:`cooksCutoff` argument for control of this behavior. Cook's distances
    for each sample are accessible as :code:`obj.layers["cooks"]`. This measure
    is useful for identifying rows where the observed counts might not fit to a
    Negative Binomial distribution.

    For analyses using the likelihood ratio test (using :func:`.nbinomLRT`),
    the p-values are determined solely by the difference in deviance between
    the full and reduced model formula. A single log2 fold change is printed in
    the results table for consistenct with other results table outputs, however
    the test statistic and p-values may nevertheless involve the testing of one
    or more log2 fold changes. Which log2 fold change is printed in the results
    table can be controlled using the :code:`name` argument, or by default this
    will be the estimated coefficient for the last element of
    :code:`obj.resultsNames()`.

    If :code:`useT = True` was specified when running :func:`.DESeq` or
    :func:`.nbinomWaldTest`, the the p-value generated by :meth:`results` will
    also make use of the t-distribution for the Wald statistic, using the
    degrees of freedom in :code:`obj.var["tDegreesFreedom"]`.

    Arguments
    ---------
    obj : DESeqDataSet
        a :class:`DESeqDataSet` one which one of the following function has
        already been called: :func:`.DESeq`, :func:`.nbinomWaldTest` or
        :func:`.nbinomLRT`.
    contrast
        this argument specifies what comparison to extract from :code:`obj` to
        build a results table. One of either:

        + a list of exactly three strings (simplest case):
            - the name of a factor in the design formula
            - the name of the numerator level for the fold change
            - the name of the denominator level for the fold change
        + a list of two string lists (more general case):
            - the names of the numerators for LFC
            - the names of the denominators for LFC
          These names should be elements of :code:`obj.resultsNames()`.
          If the list has a single element (the numerators) then the
          denominator list is considered empty.
        + a numeric list with one element for each element in
          :code:`obj.resultsNames()` (most general case)

        If specified, the :code:`name` argument is ignored.
    name : str
        the name of the individual effect (coefficient) for building a results
        table. Use this argument rahter than :code:`contrast` for continuous
        variables, individual effects or for individual interaction terms. The
        value provided to :code:`name` must be an element of
        :code:`obj.resultsNames()`.
    lfcThreshold : float
        a non-negative value which specifies a log2 fold change threshold. The
        default value is 0, corresponding to a test that the log2 fold changes
        are equal to zero. The user can specify the alternative hypothesis
        using the :code:`altHypothesis` argument, which defaults to testing for
        log2 fold changes greater in absolute value than a given threshold. If
        :code:`lfcThreshold` is specified, the results are for Wald tests, and
        LRT p-values will be overwritten.
    altHypothesis : str
        specifies the alternative hypothesis, *i.e.* those values of log2 fold
        change which the user is interested in finding. the complement of this
        set of values is the null hypothesis which will be tested. If the log2
        fold change specified by :code:`name` or by :code:`contrast` is written
        as :math:`\\beta`, then the possible values for :code:`altHypothesis`
        represent the following alternative hypotheses:

        * :code:`"greaterAbs"`: :math:`|\\beta| > \\text{lfcThreshold}`, and
          p-values are two-tailed
        * :code:`"lessAbs"`: :math:`|\\beta| < \\text{lfcThreshold}`, and
          p-values are the maximum of the upper and lower tests. The Wald
          statistic given is positive, an SE-scaled distance from the closest
          boundary
        * :code:`"greater"`: :math:`\\beta > \\text{lfcThreshold}`
        * :code:`"less"`: :math:`\\beta < \\text{lfcThreshold}`

    listValues : (float, float)
        only used if a numerators-denominators list is provided to
        :code:`contrast`.  The log2 fold changes in the list are multiplied by
        these values. The first number should be positive and the second one
        negative. Defaults to :code:`(1,-1)`.
    cooksCutoff : float
        threshold on Cook's distance, such that if one or more samples for a
        gene have a distance higher, the corresponding p-value is set to
        :code:`nan`. The default cutoff is the .99 quantile of the :math:`F(p,
        m-p)` distribution, where :math:`p` is the number of coefficients being
        fitted, and :math:`m` is the number of samples. Set to :code:`inf` or
        :code:`False` to disable the resetting of p-values to :code:`nan`.
        Note: this test excludes the Cook's distance of samples belonging to
        experimental groups with only 2 samples.
    independentFiltering : bool
        whether independent filtering should be applied automatically
    alpha : float
        the significance cutoff used for optimizing the independent filtering
        (defaults to 0.1). If the adjusted p-value cutoff (FDR) will be a value
        other than 0.1, :code:`alpha` should be set to that value.
    filter
        the vector of filter statistics over which the independent filtering
        will be optimized. By default, the mane of normalized counts is used.
    theta : array-like
        the quantiles at which to assess the number of rejections from
        independent filtering
    pAdjustMethod : str
        the method to use to adjust p-values, see :func:`.p_adjust`
    filterRun
        an optional custom function for performing independent filtering and
        p-value adjustment, with arguments :code:`res` (a
        :class:`.DESeqResults` object), :code:`filter` (the quantity for
        filtering tests), :code:`alpha` (the target FDR),
        :code:`pAdjustMethod`. This function should resturn a
        :class:`.DESeqResults` object with a :code:`padj` column.
    saveCols : array-like
        the columns of :code:`obj.var` to pass into the output results table
    test : { "Wald", "LRT" }
        automatically detected if not provided. The one exception is after
        :func:`.nbinomLRT` has been run, :code:`test="Wald"` will generate Wald
        statistics and Wald test p-values.
    addMLE : bool
        if :code:`betaPrio=True` was used (non-default). This argument
        specifies if the "unshrunken" maximum likelihood estimates (MLE) of
        log2 fold change should be added as a column to the results table
        (defaults to :code:`False`).  This argument is preserved for backward
        compatibility, as now :code:`betaPrior=True` by default and the
        recommended pipeline is to generate shrunken MAP estimates using
        :func:`lfcShrink`. This argument functionality is only implemented for
        :code:`contrast` specified as a 3-element string list.
    tidy : bool
        whether to output the results table with a header
    parallel : bool
        unimplemented
    minmu : float
        lower bound on the estimated count (used when calculating contrasts)

    Returns
    -------
    DESeqResults
        a results table (subclass of :class:`pandas.DataFrame`), containing the
        following results columns: :code:`"baseMean"`,
        :code:`"log2FoldChange"`, :code:`"lfcSE"`, :code:`"stat"`,
        :code:`"pvalue"` and :code:`"pad"`.  The :code:`"lfcSE"` gives the
        standard error of the :code:`"log2FoldChange"`.  For the Wald test,
        :code:`"stat"` is the Wald statistic: the :code:`"log2FoldChange"`
        divided by :code:`"lfcSE"`, which is compared to a standard Normal
        distribution to generate a two-tailed :code:`"pvalue"`.  For the
        likelihood ratio test (LRT), :code:`"stat"` is the difference in
        deviance between the reduced model and the full model, which is
        compared to a chi-squared distribution to generate a :code:`"pvalue"`.
    """
    if altHypothesis not in ["greaterAbs", "lessAbs", "greater", "less"]:
        raise ValueError(f"invalid value for altHypothesis: {altHypothesis}")
    if test not in [None, "Wald", "LRT"]:
        raise ValueError(f"invalid value for test: {test}")

    # initial argument testing
    if lfcThreshold < 0:
        raise ValueError("lfcThreshold should be positive")
    if alpha <= 0 or alpha >= 1:
        raise ValueError("alpha should be between 0 and 1")
    if not isinstance(listValues, (tuple, list)) or len(listValues) != 2:
        raise ValueError("listValues should be of length 2")
    if listValues[0] <= 0 or listValues[1] >= 0:
        raise ValueError("listValues should contain a positive and a negative number")
    if obj.var.type.filter(["results"]).empty:
        raise ValueError("could not find results in obj. first run DESeq()")
    if test is None:
        test = obj.test
    elif test == "Wald" and obj.test == "LRT":
        # initially test was LRT, now need to add Wald statistics and p-values
        obj = makeWaldTest(obj)
    elif test == "LRT" and obj.test == "Wald":
        raise ValueError(
            "the LRT requires the user to run nbinomLRT or DESeq(obj, test='LRT')"
        )
    if lfcThreshold == 0 and altHypothesis == "lessAbs":
        raise ValueError(
            "when testing altHypothesis='lessAbs', set the argument lfcThreshold to a positive value"
        )

    if addMLE:
        if obj.betaPrior is None:
            raise ValueError(
                "addMLE=True is only for when a beta prior was used. Otherwise, the log2 fold changes are already MLE"
            )
        if name is not None and contrast is None:
            raise ValueError(
                "addMLE=True should be used by providing string vector of length 3 to 'contrast' instead of using 'name'"
            )

    if saveCols is not None:
        try:
            obj.var[saveCols]
        except:
            raise ValueError(f"invalid value for saveCols: {saveCols}")

    hasIntercept = patsy.INTERCEPT in obj.design.design_info.terms
    isExpanded = obj.modelMatrixType == "expanded"
    noInteraction = all(len(t.factors) < 2 for t in obj.design.design_info.terms)
    # if no intercept was used or an expanded model matrix was used,
    # and neither 'contrast' nor 'name' were specified,
    # and no interactions...
    # then we create the result table: last / first level for last variable
    if (
        test == "Wald"
        and (isExpanded or not hasIntercept)
        and contrast is None
        and name is None
        and noInteraction
    ):
        designVars = [v for v in obj.design.design_info.factor_infos.values()]
        lastVarName = designVars[-1].factor.name()
        if designVars[-1].type == "categorical":
            if lastVarName.startswith("C("):
                lastVar = obj.obs[lastVarName]
            else:
                lastVar = obj.obs[f"C({lastVarName})"]
            contrast = [
                lastVarName,
                f"{lastVar.dtype.categories[-1]}",
                f"{lastVar.dtype.categories[0]}",
            ]

    if name is None:
        name = lastCoefName(obj)
    else:
        if not isinstance(name, str):
            raise ValueError("'name' should be a string")

    # done with input argument testing

    WaldResults = f"WaldPvalue_{name}" in obj.obs
    LRTResults = "LRTPValue" in obj.obs

    # this will be used in cleanContrast, and in the lfcThreshold chunks below
    useT = "tDegreesFreedom" in obj.var

    # if performing a contrast call the function cleanContrast()
    if contrast is not None:
        resNames = obj.resultsNames()
        # do some arg checking/cleaning
        contrast = checkContrast(contrast, resNames)

        ### cleanContrast call ###
        # need to go back to C++ code in order to build the beta covariance matrix
        # then this is multiplied by the numeric contrast to get the Wald statistic.
        # with 100s of samples, this can get slow, so offer parallelization
        if not parallel:
            res = cleanContrast(
                obj,
                contrast,
                expanded=isExpanded,
                listValues=listValues,
                test=test,
                useT=useT,
                minmu=minmu,
            )
        else:
            raise NotImplementedError("parallelization not implemented yet")

    else:
        # if not performing a contrast
        # pull relevant columns from obj.var
        log2FoldChange = getCoef(obj, name)
        lfcSE = getCoefSE(obj, name)
        stat = getStat(obj, name, test)
        pvalue = getPvalue(obj, name, test)
        res = pd.DataFrame(
            {
                "baseMean": obj.var["baseMean"],
                "log2FoldChange": log2FoldChange,
                "lfcSE": lfcSE,
                "stat": stat,
                "pvalue": pvalue,
            }
        )

    res.index = obj.var_names

    # add unshrunken MLE coefficients to the results table
    if addMLE:
        if not isinstance(contrast, (tuple, list)) or len(contrast) != 3:
            raise ValueError(
                "addMLE only implemented for contrast=['condition', 'B', 'A']"
            )
        res = pd.concat([res, mleContrast(obj, contrast)], axis=1)
        res = res[["baseMean", "log2FoldChange", "lfcMLE", "lfcSE", "stat", "pvalue"]]
        # if an all zero contrast, also zero out the lfcMLE
        res["lfcMLE"][(res["log2FoldChange"] == 0) & (res["stat"] == 0)] = 0

    # only if we need to generate new p-values
    if not (lfcThreshold == 0 and altHypothesis == "greaterAbs"):
        if test == "LRT":
            raise ValueError(
                "tests of log fold change above or below a threshold must be Wald test."
            )
        # check requirement if betaPrior was set to False
        if altHypothesis == "lessAbs" and obj.betaPrior:
            raise ValueError(
                "testing altHypothesis='lessAbs' requires setting the DESeq() argument betaPrior=False"
            )

        # easier to read
        LFC = res.log2FoldChange
        SE = res.lfcSE
        T = lfcThreshold

        if useT:
            df = obj.var["tDegreesFreedom"]
            pfunc = lambda q: pt(q, df=df, lower_tail=False)
        else:
            pfunc = lambda q: pnorm(q, lower_tail=False)

        if altHypothesis == "greaterAbs":
            newStat = np.sign(LFC) * np.maximum((np.abs(LFC) - T) / SE, 0)
            newPvalue = np.minimum(1, 2 * pfunc((np.abs(LFC) - T) / SE))
        elif altHypothesis == "lessAbs":
            newStatAbove = np.maximum((T - LFC) / SE, 0)
            pvalueAbove = pfunc((T - LFC) / SE)
            newStatBelow = np.maximum((LFC + T) / SE, 0)
            pvalueBelow = pfunc((LFC + T) / SE)
            newStat = np.minimum(newStatAbove, newStatBelow)
            newPvalue = np.maximum(pvalueAbove, pvalueBelow)
        elif altHypothesis == "greater":
            newStat = np.maximum((LFC - T) / SE, 0)
            newPvalue = pfunc((LFC - T) / SE)
        elif altHypothesis == "less":
            newStat = np.minimum((LFC + T) / SE, 0)
            newPvalue = pfunc((-T - LFC) / SE)

        res.stat = newStat
        res.pvalue = newPvalue

    # calculate Cook's cutoff
    m, p = obj.dispModelMatrix.shape

    if cooksCutoff is None or (isinstance(cooksCutoff, bool) and cooksCutoff):
        cooksCutoff = scipy.stats.f.ppf(0.99, p, m - p)

    # apply cutoff based on maximum Cook's distance
    # NB: cooksCutoff is not necessarily a Boolean
    performCooksCutoff = not (cooksCutoff == False)
    if performCooksCutoff:
        cooksOutlier = obj.var["maxCooks"] > cooksCutoff

        ### BEGIN heuristic to avoid filtering genes with low count outliers
        # as according to Cook's cutoff. only for two group designs.
        # do not filter if three or more counts are larger
        if np.any(cooksOutlier[~np.isnan(cooksOutlier)]):
            designVars = obj.design.design_info.factor_infos
            if len(designVars) == 1:
                var = [v for v in designVars.values()][0]
                if var.type == "categorical" and len(var.categories) == 2:
                    maxIndices = np.argmax(obj.layers["cooks"][:, cooksOutlier], axis=0)
                    outliers = obj.counts()[:, cooksOutlier]
                    # counts for the outliers with max cooks
                    outCount = np.take_along_axis(
                        outliers, maxIndices[None], 0
                    ).squeeze()
                    # if three or more counts larger than the outlier
                    # do not filter out the p-value for those genes
                    dontFilter = np.sum(outliers > outCount, axis=0) >= 3
                    # reset the outlier status for these genes
                    cooksOutlier[cooksOutlier] &= ~dontFilter
        ### END heuristic

        res.pvalue[cooksOutlier] = np.nan

    # if original baseMean was positive, but now zero due to replaced counts,
    # fill in results
    if "replace" in obj.var and np.nansum(obj.var["replace"]) > 0:
        nowZero = obj.var["replace"] & (obj.var["baseMean"] == 0)
        res.log2FoldChange[nowZero] = 0
        if addMLE:
            res.lfcMLE[nowZero] = 0
        res.lfcSE[nowZero] = 0
        res.stat[nowZero] = 0
        res.pvalue[nowZero] = 1

    # add prior information
    if not obj.betaPrior:
        priorInfo = {
            "type": "none",
            "package": "inmoose.deseq2",
            "version": __version__,
        }
    else:
        priorInfo = {
            "type": "normal",
            "package": "inmoose.deseq2",
            "version": __version__,
            "betaPriorVar": obj.betaPriorVar,
        }

    # make results object
    deseqRes = DESeqResults(res, priorInfo=priorInfo)

    # p-value adjustment
    if filterRun is None:
        deseqRes = pvalueAdjustment(
            deseqRes, independentFiltering, filter, theta, alpha, pAdjustMethod
        )
    else:
        deseqRes = filterRun(deseqRes, filter, alpha, pAdjustMethod)

    # stash lfcThreshold
    deseqRes.lfcThreshold = lfcThreshold

    # remove rownames and attach a new column 'row'
    if tidy:
        raise NotImplementedError()

    if saveCols is not None:
        deseqRes.mrows2Save = obj.var[saveCols]

    return deseqRes


def lastCoefName(obj):
    """convenience function to guess the name of the last coefficient
    in the model matrix, unless specified this will be used for plots and
    accessor functions.
    """
    return obj.resultsNames()[-1]


def getCoef(obj, name):
    if name is None:
        name = lastCoefName(obj)
    return obj.var[name]


def getCoefSE(obj, name):
    if name is None:
        name = lastCoefName(obj)
    return obj.var[f"SE_{name}"]


def getStat(obj, name, test="Wald"):
    if name is None:
        name = lastCoefName(obj)

    if test == "Wald":
        return obj.var[f"WaldStatistic_{name}"]
    elif test == "LRT":
        return obj.var["LRTStatistic"]
    else:
        raise ValueError(f"unknown test: {test}")


def getPvalue(obj, name, test="Wald"):
    if name is None:
        name = lastCoefName(obj)

    if test == "Wald":
        return obj.var[f"WaldPvalue_{name}"]
    elif test == "LRT":
        return obj.var["LRTPvalue"]
    else:
        raise ValueError(f"unknown test: {test}")


def pvalueAdjustment(res, independentFiltering, filter, theta, alpha, pAdjustMethod):
    # perform independent filtering
    if independentFiltering:
        if filter is None:
            filter = res.baseMean
        if theta is None:
            lowerQuantile = np.mean(filter == 0)
            if lowerQuantile < 0.95:
                upperQuantile = 0.95
            else:
                upperQuantile = 1.0
            theta = np.linspace(lowerQuantile, upperQuantile, 50)

        # do filtering using genefilter
        if len(theta) <= 1:
            raise ValueError("theta should be a list")
        if len(filter) != res.shape[0]:
            raise ValueError("filter should have as many elements as res has rows")
        filtPadj = filtered_p(
            filter_=filter, test=res.pvalue, theta=theta, method=pAdjustMethod
        )
        numRej = np.nansum(filtPadj < alpha, axis=0)
        # prevent over-aggressive filtering when all genes are null,
        # by requiring the max number of rejections is above a fitted curve.
        # If the max number of rejection is not greater than 10, then don't
        # perform independent filtering at all.
        lo_fit = lowess(numRej, theta, frac=1 / 5)
        if np.max(numRej) <= 10:
            j = 0
        else:
            if np.all(numRej == 0):
                residual = 0
            else:
                residual = numRej[numRej > 0] - lo_fit[numRej > 0, 1]
            thresh = np.max(lo_fit[:, 1]) - np.sqrt(np.mean(residual**2))
            if np.any(numRej > thresh):
                j = np.nonzero(numRej > thresh)[0][0]
            else:
                j = 0

        padj = filtPadj[:, j]
        cutoffs = np.quantile(filter, theta)
        filterThreshold = cutoffs[j]
        filterNumRej = pd.DataFrame({"theta": theta, "numRej": numRej})
        filterTheta = theta[j]

        res.filterThreshold = filterThreshold
        res.filterTheta = filterTheta
        res.filterNumRej = filterNumRej
        res.lo_fit = lo_fit
        res.alpha = alpha

    else:
        # regular p-value adjustment
        # does not include those rows which were removed
        # by maximum Cook's distance
        padj = p_adjust(res.pvalue, method=pAdjustMethod)

    res["padj"] = padj
    res.type["padj"] = "results"
    res.description["padj"] = f"{pAdjustMethod} adjusted p-values"

    return res


def filtered_p(filter_, test, theta, method, data=None):
    """
    compute and adjust p-values, with filtering

    Given filter and test statistics in the form of unadjusted p-values, or
    functions able to compute these statistics from the data, filter and then
    correct the p-values across a range of filtering stringencies.

    NB: this function is copied and ported from the R package `genefilter
    <https://bioconductor.org/packages/release/bioc/html/genefilter.html>`_.
    This copy was already done in the original DESeq2 R code.

    Arguments
    ---------
    filter_
        a list of stage-one filter statistics, or a function which is able to
        compute such a list from :code:`data`, if :code:`data` is supplied
    test
        a list of unadjusted p-values, or a function which is able to compute
        such a list from the filtered portions of :code:`data`, if :code:`data`
        is supplied. The option to supply a function is useful when the value
        of the test statistic depends on which hypotheses are filtered out at
        stage one.
    theta
        a list with one or more filtering fractions to consider. Actual cutoffs
        are then computed internally by applying :code:`quantile` to the filter
        statistics contained in (or produced by) the :code:`filter_` argument.
    method
        the unadjusted p-values contained in (or produced by) :code:`test` will
        be adjusted for multiple testing after filtering, using
        :func:`p_adjust`. See the :code:`method` argument of this function for
        more options.
    data
        if :code:`filter_` and/or :code:`test` are functions rather than lists
        of statistics, they will be applied to :code:`data`. The functions will
        be passed the whole :code:`data` object, and must work over rows etc.
        themselves as appropriate.

    Returns
    -------
    ndarray
        a matrix of p-values, possibly adjusted for multiple testing, with one
        row per  null hypothesis and one column per filtering fraction given in
        :code:`theta`. For a given column, entries which have been filtered out
        are :code:`nan`.
    """
    if callable(filter_):
        U1 = filter_(data)
    else:
        U1 = filter_

    cutoffs = np.quantile(U1, theta)
    result = np.full((len(U1), len(cutoffs)), np.nan)
    for i in range(len(cutoffs)):
        use = U1 >= cutoffs[i]
        if np.any(use):
            if callable(test):
                U2 = test(data[use, :])
            else:
                U2 = test[use]
            result[use, i] = p_adjust(U2, method=method)
    return result


def getContrast(obj, contrast, useT, minmu):
    """takes a DESeqDataSet obj and a numeric vector specifying a contrast
    and returns a vector of Wald statistics corresponding to the contrast.
    """
    if contrast is None:
        raise ValueError("must provide a contrast")
    modelMatrix = obj.modelMatrix

    # only continue on the cols with non-zero col mean
    objNZ = obj[:, ~obj.var["allZero"]]
    normalizationFactors = objNZ.getSizeOrNormFactors()
    alpha_hat = objNZ.var["dispersion"]
    # convert beta to log scale
    beta_mat = np.log(2) * objNZ.var.description.filter(regex="log2 fold change")
    # convert beta prior variance to log scale
    lambda_ = 1 / (np.log(2) ** 2 * obj.betaPriorVar)

    # check if DESeq() replaced outliers
    if "replaceCounts" in obj.layers:
        countsMatrix = objNZ.layers["replaceCounts"]
    else:
        countsMatrix = objNZ.counts()

    # use weights if they are present
    if "weights" in obj.layers:
        useWeights = True
        weights = objNZ.layers["weights"]
        if not np.all(weights >= 0):
            raise ValueError("all weights must be positive")
        weights = weights / np.max(weights, axis=0)
    else:
        useWeights = False
        weights = np.ones(objNZ.shape)

    betaRes = fitBeta(
        y=countsMatrix,
        x=modelMatrix,
        nf=normalizationFactors,
        alpha_hat=alpha_hat.values,
        contrast=contrast,
        beta_mat=beta_mat.values,
        lambda_=lambda_,
        weights=weights,
        useWeights=useWeights,
        tol=1e-8,
        maxit=0,
        useQR=False,  # QR not relevant, fitting loop is not entered
        minmu=minmu,
    )
    # convert back to log2 scale
    contrastEstimate = np.log2(np.exp(1)) * betaRes["contrast_num"]
    contrastSE = np.log2(np.exp(1)) * betaRes["contrast_denom"]
    contrastStatistic = contrastEstimate / contrastSE

    if useT:
        if "tDegreesFreedom" not in obj.var:
            raise ValueError("tDegreesFreedom should be in obj.var")
        df = objNZ.var["tDegreesFreedom"]
        contrastPvalue = 2 * pt(np.abs(contrastStatistic), df=df, lower_tail=False)
    else:
        contrastPvalue = 2 * pnorm(np.abs(contrastStatistic), lower_tail=False)

    contrastResults = pd.DataFrame(
        {
            "log2FoldChange": contrastEstimate,
            "lfcSE": contrastSE,
            "stat": contrastStatistic,
            "pvalue": contrastPvalue,
        }
    )
    contrastResults = buildDataFrameWithNACols(contrastResults, obj.var["allZero"])
    contrastResults.index = obj.var_names
    return contrastResults


def cleanContrast(obj, contrast, expanded, listValues, test, useT, minmu):
    """this function takes a desired contrast as specified by results(),
    performs checks, and then either returns the already existing contrast
    or generates the contrast by calling getContrast() using a numeric vector
    """
    # get the names of columns in the beta matrix
    resNames = obj.resultsNames()
    # if possible, return pre-computed columns, which are
    # already stored in obj.var. This will be the case using
    # results() with 'name', or if expanded model matrices were not
    # run and the contrast contains the reference level as numerator and denominator

    resReady = False

    if all(isinstance(c, str) for c in contrast):
        contrastFactor = contrast[0]
        if not contrastFactor.startswith("C("):
            contrastFactor = f"C({contrastFactor})"
        contrastFactorName = getFactorName(contrastFactor)
        if contrastFactorName not in obj.obs:
            raise ValueError(
                f"{contrastFactorName} should be the name of a factor in the obs data of the DESeqDataSet"
            )
        if contrastFactor not in obj.obs or not pd.api.types.is_categorical_dtype(
            obj.obs[contrastFactor]
        ):
            raise ValueError(f"{contrastFactorName} is not a factor")

        contrastNumLevel = contrast[1]
        contrastDenomLevel = contrast[2]
        # make sure contrastBaseLevel is a string
        contrastBaseLevel = f"{obj.obs[contrastFactor].dtype.categories[0]}"

        # check for intercept
        hasIntercept = 0 in [len(t.factors) for t in obj.design.design_info.terms]
        firstVar = (
            contrast[0]
            == [t.name() for t in obj.design.design_info.terms if len(t.factors) > 0][0]
        )

        # tricky case: if the design has no intercept, the factor is not the
        # first variable in the design, and one of the numerator or
        # denominator is the reference level, then the desired contrast is
        # simply a coefficient (or -1 times)
        noInterceptPullCoef = (
            not hasIntercept
            and not firstVar
            and (contrastBaseLevel in [contrastNumLevel, contrastDenomLevel])
        )

        # case 1: standard model matrices: pull coef or build the appropriate contrast
        # coefficients names are of the form "factor_level_vs_baselevel"
        # output: contrastNumColumn and contrastDenomColumn
        if not expanded and (hasIntercept or noInterceptPullCoef):
            contrastNumColumn = (
                f"{contrastFactorName}_{contrastNumLevel}_vs_{contrastBaseLevel}"
            )
            contrastDenomColumn = (
                f"{contrastFactorName}_{contrastDenomLevel}_vs_{contrastBaseLevel}"
            )
            # check that the desired contrast is already available in obj.var,
            # and then we can either take it directly or multiply the log fold
            # change and Wald stat by -1
            if contrastDenomLevel == contrastBaseLevel:
                cleanName = (
                    f"{contrastFactorName} {contrastNumLevel} vs {contrastDenomLevel}"
                )
                # the results can be pulled directly from obj.var
                if hasIntercept or not firstVar:
                    name = f"{contrastFactorName}_{contrastNumLevel}_vs_{contrastDenomLevel}"
                else:
                    name = f"{contrastFactorName}[{contrastNumLevel}]"

                if name not in resNames:
                    raise ValueError(
                        f"as {contrastDenomLevel} is the reference level, was expecting {name} to be present in 'resultsNames(obj)'"
                    )

                log2FoldChange = getCoef(obj, name)
                lfcSE = getCoefSE(obj, name)
                stat = getStat(obj, name, test)
                pvalue = getPvalue(obj, name, test)
                res = pd.DataFrame(
                    {
                        "baseMean": obj.var["baseMean"],
                        "log2FoldChange": log2FoldChange,
                        "lfcSE": lfcSE,
                        "stat": stat,
                        "pvalue": pvalue,
                    }
                )
                lfcType = "MAP" if obj.betaPrior else "MLE"
                res.description[
                    "log2FoldChange"
                ] = f"log2 fold change ({lfcType}): {cleanName}"
                resReady = True

            elif contrastNumLevel == contrastBaseLevel:
                # fetch the results for denom vs num
                # and multiply the log fold change and stat by -1
                cleanName = (
                    f"{contrastFactorName} {contrastNumLevel} vs {contrastDenomLevel}"
                )
                if hasIntercept or not firstVar:
                    swapName = f"{contrastFactorName}_{contrastDenomLevel}_vs_{contrastNumLevel}"
                else:
                    swapName = f"{contrastFactorName}[{contrastDenomLevel}]"

                if swapName not in resNames:
                    raise ValueError(
                        f"as {contrastNumLevel} is the reference level, was expecting {swapName} to be present in resultsNames(obj)"
                    )

                log2FoldChange = getCoef(obj, swapName)
                lfcSE = getCoefSE(obj, swapName)
                stat = getStat(obj, swapName, test)
                pvalue = getPvalue(obj, swapName, test)
                res = pd.DataFrame(
                    {
                        "baseMean": obj.var["baseMean"],
                        "log2FoldChange": log2FoldChange,
                        "lfcSE": lfcSE,
                        "stat": stat,
                        "pvalue": pvalue,
                    }
                )
                res["log2FoldChange"] *= -1
                if test == "Wald":
                    res["stat"] *= -1
                lfcType = "MAP" if obj.betaPrior else "MLE"
                res.description[
                    "log2FoldChange"
                ] = f"log2 fold change ({lfcType}): {cleanName}"
                res.description["lfcSE"] = f"standard error: {cleanName}"
                # rename some of the columns using the flipped contrast
                if test == "Wald":
                    res.description["stat"] = f"Wald statistic: {cleanName}"
                    res.description["pvalue"] = f"Wald test p-value: {cleanName}"

                resReady = True

            else:
                # check for the case where neither are present
                # as comparisons against reference level
                if not (
                    contrastNumColumn in resNames and contrastDenomColumn in resNames
                ):
                    raise ValueError(
                        f"{contrastNumLevel} and {contrastDenomLevel} should be levels of {contrastFactorName} such that {contrastNumLevel} and {contrastDenomLevel} are contained in 'resultsNames(obj)'"
                    )

        # case 2: expanded model matrices or no intercept and first variable
        # need to then build the appropriate contrast.
        # these coefficient names have the form "factorLevel"
        # output: contrastNumColumn and contrastDenomColumn
        else:
            # we only need to check validity
            if hasIntercept:
                contrastNumColumn = f"{contrastFactorName}[T.{contrastNumLevel}]"
                contrastDenomColumn = f"{contrastFactorName}[T.{contrastDenomLevel}]"
            elif not firstVar:
                contrastNumColumn = (
                    f"{contrastFactorName}_{contrastNumLevel}_vs_{contrastBaseLevel}"
                )
                contrastDenomColumn = (
                    f"{contrastFactorName}_{contrastDenomLevel}_vs_{contrastBaseLevel}"
                )
            else:
                contrastNumColumn = f"{contrastFactorName}[{contrastNumLevel}]"
                contrastDenomColumn = f"{contrastFactorName}[{contrastDenomLevel}]"
            if not (contrastNumColumn in resNames and contrastDenomColumn in resNames):
                raise ValueError(
                    f"{contrastNumColumn} and {contrastDenomColumn} are expected to be in resultsNames(obj): {resNames}"
                )

        # check if both level have all zero counts
        # (this has to be done here to make use of error checking above)
        contrastAllZero = contrastAllZeroCharacter(
            obj, contrastFactor, contrastNumLevel, contrastDenomLevel
        )

    # if the result table not already built in the above code
    if not resReady:
        # here, a numeric / list / string contrast which will be converted
        # into a numeric contras and run through getContrast()
        if all(isinstance(c, (int, float, np.number)) for c in contrast):
            # make name for numeric contrast
            signMap = ["", "", "+"]
            contrastSigns = [signMap[x + 1] for x in np.sign(contrast)]
            contrastName = ",".join(
                [f"{s}{c}" for s, c in zip(contrastSigns, contrast)]
            )
            # make sure the contrast is an np array
            contrast = np.asarray(contrast)
        elif all(isinstance(c, str) for c in contrast):
            # interpret string contrast into numeric and make a name for the contrast
            contrastNumeric = np.zeros(len(resNames))
            contrastNumeric[resNames == contrastNumColumn] = 1
            contrastNumeric[resNames == contrastDenomColumn] = -1
            contrast = contrastNumeric
            contrastName = (
                f"{contrastFactorName} {contrastNumLevel} vs {contrastDenomLevel}"
            )
        else:
            # interpret 2-list contrast into numeric and make a name for the contrast
            lc1 = len(contrast[0])
            lc2 = len(contrast[1])
            # these just used for naming
            listvalname1 = round(listValues[0], 3)
            listvalname2 = round(listValues[1], 3)
            if lc1 > 0 and lc2 > 0:
                listvalname2 = np.abs(listvalname2)
                listvalname1 = "" if listvalname1 == 1 else f"{listvalname1} "
                listvalname2 = "" if listvalname2 == 1 else f"{listvalname2} "
                contrastName = f"{listvalname1}{'+'.join(contrast[0])} vs {listvalname2}{'+'.join(contrast[1])}"
            elif lc1 > 0 and lc2 == 0:
                listvalname1 = "" if listvalname1 == 1 else f"{listvalname1} "
                contrastName = f"{listvalname1}{'+'.join(contrast[0])} effect"
            elif lc1 == 0 and lc2 > 0:
                contrastName = f"{listvalname2}{'+'.join(contrast[1])} effect"

            contrastNumeric = np.zeros(len(resNames))
            contrastNumeric[resNames.isin(contrast[0])] = listValues[0]
            contrastNumeric[resNames.isin(contrast[1])] = listValues[1]
            contrast = contrastNumeric

        contrastAllZero = contrastAllZeroNumeric(obj, contrast)

        # now get the contrast
        res = getContrast(obj, contrast, useT=useT, minmu=minmu)
        lfcType = "MAP" if obj.betaPrior else "MLE"
        for c in res.columns:
            res.type[c] = "results"
        res.description[
            "log2FoldChange"
        ] = f"log2 fold change ({lfcType}): {contrastName}"
        res.description["lfcSE"] = f"standard error: {contrastName}"
        res.description["stat"] = f"Wald statistic: {contrastName}"
        res.description["pvalue"] = f"Wald test p-value: {contrastName}"
        res["baseMean"] = obj.var["baseMean"]
        res.type["baseMean"] = obj.var.type["baseMean"]
        res.description["baseMean"] = obj.var.description["baseMean"]

    # if the counts in all samples included in contrast are zero
    # then zero out the LFC, Wald stat and p-value set to 1
    contrastAllZero = contrastAllZero & ~obj.var["allZero"]
    contrastAllZero.index = res.index
    if np.sum(contrastAllZero) > 0:
        res.log2FoldChange[contrastAllZero] = 0
        res.stat[contrastAllZero] = 0
        res.pvalue[contrastAllZero] = 1

    # if test is "LRT", overwrite the statistic and p-value
    # (we only ran contrast for the coefficient)
    if test == "LRT":
        stat = getStat(obj, name=None, test=test)
        pvalue = getPvalue(obj, name=None, test=test)
        res = res[["baseMean", "log2FoldChange", "lfcSE"]]
        res["stat"] = stat
        res["pvalue"] = pvalue

    return res


def mleContrast(obj, contrast):
    contrastFactor = contrast[0]
    contrastNumLevel = contrast[1]
    contrastDenomLevel = contrast[2]
    contrastRefLevel = obj.obs[contrastFactor].dtype.categories[0]
    contrastNumColumn = f"MLE_{contrastFactor}_{contrastNumLevel}_vs_{contrastRefLevel}"
    contrastDenomColumn = (
        f"MLE_{contrastFactor}_{contrastDenomLevel}_vs_{contrastRefLevel}"
    )
    cleanName = f"log2 fold change (MLE): {contrastFactor} {contrastNumLevel} vs {contrastDenomLevel}"

    if contrastDenomLevel == contrastRefLevel:
        name = f"MLE_{contrastFactor}_{contrastNumLevel}_vs_{contrastDenomLevel}"
        lfcMLEName = name
        lfcMLE = obj.var[name]
    elif contrastNumLevel == contrastRefLevel:
        swapName = f"MLE_{contrastFactor}_{contrastDenomLevel}_vs_{contrastNumLevel}"
        lfcMLEName = swapName
        lfcMLE = -1 * obj.var[swapName]
    else:
        numMLE = obj.var[contrastNumColumn]
        denomMLE = obj.var[contrastDenomColumn]
        lfcMLEName = contrastNumColumn
        lfcMLE = numMLE - denomMLE

    res = pd.DataFrame({"lfcMLE": lfcMLE})
    res.type["lfcMLE"] = obj.var.type[lfcMLEName]
    res.description["lfcMLE"] = cleanName
    return res


def is_iterable(x):
    try:
        return not isinstance(x, str) and all(True for y in x)
    except TypeError:
        return False


def checkContrast(contrast, resNames):
    if not is_iterable(contrast):
        contrast = [contrast]
    is_numeric = all(isinstance(c, (int, float, np.number)) for c in contrast)
    is_string = all(isinstance(c, str) for c in contrast)

    if is_string and len(contrast) == 3:
        if contrast[1] == contrast[2]:
            raise ValueError(
                f"{contrast[1]} and {contrast[2]} should be different level names"
            )

    elif is_numeric:
        if len(contrast) != len(resNames):
            raise ValueError(
                "numeric contrast vector should have one element for every element of 'resultsNames(obj)'"
            )
        if np.all(np.equal(contrast, 0)):
            raise ValueError(
                "numeric contrast vector cannot have all elements equal to 0"
            )
    else:
        if len(contrast) == 1:
            contrast = [contrast[0], []]
        if len(contrast) != 2:
            raise ValueError(
                "'contrast', as a pair of lists, should have length 2, or if length 1, an empty list will be added as the second element"
            )

        # make sure contrast is a pair of iterables
        contrast = [c if is_iterable(c) else [c] for c in contrast]

        if not all(isinstance(c, str) for cc in contrast for c in cc):
            raise ValueError(
                "'contrast', as a pair of lists, should have lists of strings as elements"
            )
        if not all(c in resNames for cc in contrast for c in cc):
            raise ValueError(
                "all elements of the 2-element contrast should be elements of 'resultsNames(obj)'"
            )
        if len(np.intersect1d(contrast[0], contrast[1])) > 0:
            raise ValueError(
                "elements in the 2-element contrast should only appear in the numerator (first element of contrast) or the denominator(second element), but not both"
            )
        if len(contrast[0]) + len(contrast[1]) == 0:
            raise ValueError(
                "one the 2 elements in the list should be contain at least one element"
            )

    return contrast


def contrastAllZeroCharacter(obj, contrastFactor, contrastNumLevel, contrastDenomLevel):
    cts = obj.counts()
    f = obj.obs[contrastFactor]
    if np.issubdtype(f.dtype.categories.dtype, np.number):
        contrastNumLevel = int(contrastNumLevel)
        contrastDenomLevel = int(contrastDenomLevel)
    cts_sub = cts[f.isin([contrastNumLevel, contrastDenomLevel])]
    return np.sum(cts_sub == 0, axis=0) == cts_sub.shape[0]


def contrastAllZeroNumeric(obj, contrast):
    modelMatrix = obj.modelMatrix
    # this extra leg-work to zero out FLC, lfcSE, and set p-value to 1
    # for contrasts comparing groups where both groups have all zeros.

    # it is only implemented for the case in which we can identify
    # the relevant samples by multiplying the model matrix
    # with a vector where the non-zero elements of the numeric contrast
    # are replaced with 1.

    # so, this code will not zero out in the case of standard model matrices
    # where the user supplies a numeric contrast that pulls out a single column
    # of the model matrix, for example.

    if np.all(np.greater_equal(contrast, 0)) or np.all(np.less_equal(contrast, 0)):
        return np.repeat(False, obj.n_vars)

    contrastBinary = np.where(contrast == 0, 0, 1)
    whichSamples = np.where(modelMatrix @ contrastBinary == 0, 0, 1)
    zeroTest = obj.counts().T @ whichSamples
    return zeroTest == 0



# -----------------------------------------------------------------------------
# Copyright (C) 2008-2022 Yunshun Chen, Aaron TL Lun, Davis J McCarthy, Matthew E Ritchie, Belinda Phipson, Yifang Hu, Xiaobei Zhou, Mark D Robinson, Gordon K Smyth
# Copyright (C) 2022-2023 Maximilien Colange

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
# -----------------------------------------------------------------------------

# This file is based on the file 'R/DGEList.R' of the Bioconductor edgeR package (version 3.38.4).


import numpy as np


def _isAllZero(y):
    """
    Check for all-zero, negative, Nan and infinite counts

    This function check if :code:`y` is all zero, and raises an error if it
    contains negative, NaN or infinite values.

    Arguments
    ---------
    y : array_like
        matrix of counts

    Returns
    -------
    bool
        whether :code:`y` only contains zeroes
    """
    if len(y) == 0:
        return False
    check_range = (np.amin(y), np.nanmax(y))
    if np.isnan(check_range[0]):
        raise ValueError("NaN counts are not allowed")
    if check_range[0] < 0:
        raise ValueError("negative counts are not allowed")
    if np.isinf(check_range[1]):
        raise ValueError("infinite counts are not allowed")
    return check_range[1] == 0



# -----------------------------------------------------------------------------
# Copyright (C) 2013-2022 Michael I. Love, Constantin Ahlmann-Eltze
# Copyright (C) 2023 Maximilien Colange

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
# -----------------------------------------------------------------------------

# This file is based on the file 'R/core.R' of the Bioconductor DESeq2 package
# (version 3.16).


import logging
import numpy as np
from scipy.special import polygamma
from scipy.stats import trim_mean
import statsmodels.api as sm
from statsmodels.tools.sm_exceptions import DomainWarning

from ..utils import Factor
from .fitNbinomGLMs import fitNbinomGLMs
from .misc import checkFullRank, buildMatrixWithNACols, buildVectorWithNACols
from .weights import getAndCheckWeights
from .deseq2_cpp import fitDispWrapper, fitDispGridWrapper


def estimateDispersions_dds(
    obj,
    fitType="parametric",
    maxit=100,
    useCR=True,
    weightThreshold=1e-2,
    quiet=False,
    modelMatrix=None,
    minmu=None,
):
    """
    Estimate the dispersions for a :class:`DESeqDataSet`

    This function obtains dispersion estimates for Negative Binomial
    distributed data. The function is typically called with the idiom:
    :code:`dds = dds.estimateDispersions()`.

    The fitting proceeds as follows: for each gene, an estimate of the
    dispersion is found which maximizes the Cox-Reid adjusted profile
    likelihood (the methods of Cox-Reid adjusted profile likelihood
    maximization for estimation of dispersion in RNA-Seq data were developed by
    [McCarthy2012]_, first implemented in the edgeR package in 2010 (see
    :func:`inmoose.edgepy.glmFit` and :func:`inmoose.edgepy.dispCoxReid`); a
    trend line capturing the dispersion-mean relationship is fit to the maximum
    likelihood estimates; a normal prior is determined for the log dispersion
    estimates centered on the predicted value from the trended fit with
    variance equal to the difference between the observed variance of the log
    dispersion estimates and te expected sampling variance; finally maximum a
    posteriori dispersion estimates are returned.  This final dispersion
    parameter is used in subsequent tests. The final dispersion estimates can
    be accessed through :attr:`DESeqDataSet.dispersions`. The fitted
    dispersion-mean relationship is also used in
    :func:`varianceStabilizingTransformation`. All of the intermediate values
    (gene-wise dispersion estimates, fitted dispersion estimates from the
    trended fit, etc.) are stored in :code:`dds.var`.

    The log normal prior on the dispersion parameter has been proposed by [Wu2012]_.

    See also
    --------
    ~inmoose.deseq2.estimateDispersionsGeneEst
        lower-level function called by this function
    ~inmoose.deseq2.estimateDispersionsFit
        lower-level function called by this function
    ~inmoose.deseq2.estimateDispersionsMAP
        lower-level function called by this function

    Arguments
    ---------
    obj : DESeqDataSet
        the input dataset
    fitType : "parametric", "local", "mean" or "glmGamPoi"
        the type of fitting the dispersions to the mean intensity.

        :code:`"parametric"` - fit a dispersion-mean relation of the form:
        :math:`dispersion = asymtDisp + extraPois / mean` *via* a robust
        gamma-family GLM.  The coefficients :code:`asymtDisp` and
        :code:`extraPois` are given in the attribute :code:`coefficients` of
        the :attr:`DESeqDataSet.dispersionFunction`.

        :code:`"local"` - use the locfit package to fit a local regression of
        log dispersions over log base mean (normal scale means and dispersions
        are input and output for :attr:`DESeqDataSet.dispersionFunction`. The
        points are weighted by normalized mean count in the local regression.

        :code:`"mean"` - use the mean of gene-wise dispersion estimates.

        :code:`"glmGamPoi"` - use the glmGamPoi package to fit the gene-wise
        dispersion, its trend, and calculate the MAP based on the
        quasi-likelihood framework. The trend is calculated using a local
        median regression.

    maxit : int
        maximum number of iterations to allow for convergence
    useCR : bool
        whether to use Cox-Reid adjustment (see [McCarthy2012]_)
    weightThreshold : float
        threshold for subsetting the design matrix and GLM weights for
        calculating the Cox-Reid correction
    quiet : bool
        whether to print messages at each step
    modelMatrix : array-like, optional
        an optional matrix which will be used for fitting the expected counts.
        By default, the model matrix is constructed from
        :attr:`DESeqDataSet.design`.
    minmu : float
        lower bound on the estimated count for fitting gene-wise dispersion

    Returns
    -------
    DESeqDataSet
        the input :code:`obj`, with the dispersion information filled in fields
        in :attr:`DESeqDataSet.var`, or the final dispersions accessible via
        :attr:`DESeqDataSet.dispersions`.
    """

    if fitType not in ["parametric", "local", "mean", "glmGamPoi"]:
        raise ValueError(f"invalid value for fitType: {fitType}")
    if minmu is None:
        if fitType == "glmGamPoi":
            minmu = 1e-6
        else:
            minmu = 0.5

    if obj.sizeFactors is None and obj.normalizationFactors is None:
        raise ValueError(
            "first call estimateSizeFactors or provide a normalizationFactor matrix before calling estimateDispersions"
        )

    # size factors could have slipped in the obs from a previous run
    if obj.sizeFactors is not None:
        if not np.issubdtype(obj.sizeFactors.dtype, np.number):
            raise ValueError(
                "the sizeFactor column in obs is not numeric. This column could have come in during obs import and should be removed."
            )
        if np.isnan(obj.sizeFactors).any():
            raise ValueError(
                "the sizeFactor column in obs contains NA. This column could have come in during obs import and should be removed."
            )

    if (np.sum(obj.X == obj.X[0], 1) == obj.n_obs).all():
        raise ValueError(
            "all genes have equal values for all samples. will not be able to perform differential analysis"
        )

    if "dispersion" in obj.var:
        logging.info("found already estimated dispersions, replacing these")
        del obj.var["dispersion"]

    if fitType == "glmGamPoi":
        dispersionEstimator = "glmGamPoi"
        raise NotImplementedError(
            "estimating dispersions with glmGamPoi is not implemented"
        )
    else:
        dispersionEstimator = "DESeq2"

    checkForExperimentalReplicates(obj, modelMatrix)

    logging.info("gene-wise dispersion estimates")
    obj = estimateDispersionsGeneEst(
        obj,
        maxit=maxit,
        useCR=useCR,
        weightThreshold=weightThreshold,
        quiet=quiet,
        modelMatrix=modelMatrix,
        minmu=minmu,
        type_=dispersionEstimator,
    )
    logging.info("mean-dispersion relationship")
    obj = estimateDispersionsFit(obj, fitType=fitType, quiet=quiet)
    logging.info("final dispersion estimates")
    obj = estimateDispersionsMAP(
        obj,
        maxit=maxit,
        useCR=useCR,
        weightThreshold=weightThreshold,
        quiet=quiet,
        modelMatrix=modelMatrix,
        type_=dispersionEstimator,
    )
    return obj


def estimateDispersionsGeneEst(
    obj,
    minDisp=1e-8,
    kappa_0=1,
    dispTol=1e-6,
    maxit=100,
    useCR=True,
    weightThreshold=1e-2,
    quiet=False,
    modelMatrix=None,
    niter=1,
    linearMu=None,
    minmu=None,
    alphaInit=None,
    type_="DESeq2",
):
    """
    Low-level function to fit dispersion estimates

    Normal users should instead use :meth:`.DESeqDataSet.estimateDispersions`.
    This low-level function is called by
    :meth:`.DESeqDataSet.estimateDispersions`, but is exported and documented
    for non-standard usage.  For instance, it is possible to replace fitted
    values with a custom fit and continue with the maximum a posteriori
    dispersion estimate.

    Arguments
    ---------
    obj : DESeqDataSet
        the input dataset
    minDisp : float
        small value for the minimum dispersion, to allow for calculations in
        log scale, one order of magnitude above this value is used as a test
        for inclusion in mean-dispersion fitting
    kappa_0 : float
        parameter used in setting the initial proposal in backtracking search,
        higher :code:`kappa_0` results in larger steps
    dispTol : float
        parameter to test for convergence of log dispersion, stop when increase
        in log posterior is less than :code:`dispTol`
    maxit : int
        maximum number of iterations to allow for convergence
    useCR : bool
        whether to use Cox-Reid adjustment
    weightThreshold : float
        threshold for subsetting the design matrix and GLM weights for
        calculating the Cox-Reid correction
    quiet : bool
        whether to print messages at each step
    modelMatrix : array-like
        for advanced use only, a substitute model matrix for gene-wise and MAP
        dispersion estimation
    niter : int
        number of times to iterate between estimation of means and estimation
        of dispersion
    linearMu :
        estimate the expected counts matrix using a linear model.
        Defaults to :code:`None`, in which case a linear model is used if the
        number of groups defined by the model matrix is equal to the number of
        columns of the model matrix.
    minmu : float
        lower bound on the estimated count for fitting gene-wise dispersion
    alphaInit : array-like
        initial guess for the dispersion estimates
    type_ : "DESeq2" or "glmGamPoi"
        specify if the glmGamPoi package is used to calculate the dispersion.
        This can be significantly faster if there are many replicates with
        small counts.

    Returns
    -------
    DESeqDataSet
        the input :code:`obj` with gene-wise dispersion estimates
    """
    if type_ not in ["DESeq2", "glmGamPoi"]:
        raise ValueError(f"invalid value for type_: {type_}")
    if minmu is None:
        if type_ == "glmGamPoi":
            minmu = 1e-6
        else:
            minmu = 0.5

    if "dispGeneEst" in obj.var:
        logging.info("found already estimated gene-wise dispersions, removing these")
        del obj.var["dispGeneEst"]
        if "dispGeneIter" in obj.var:
            del obj.var["dispGeneIter"]

    if np.log(minDisp / 10) <= -30:
        raise ValueError(
            "for computational stability, log(minDisp/10) should be above -30"
        )

    if modelMatrix is None:
        modelMatrix = obj.design
    checkFullRank(modelMatrix)
    if modelMatrix.shape[0] == modelMatrix.shape[1]:
        raise ValueError(
            "the number of samples and the number of model coefficients are equal, i.e., there are no replicates to estimate the dispersion. Use an alternate design formula."
        )

    obj = obj.getBaseMeansAndVariances()

    # use weights if they are present
    # (we need this already to decide about linear mu fitting)
    obj.weightsOK = None
    (obj, weights, useWeights) = getAndCheckWeights(
        obj, modelMatrix, weightThreshold=weightThreshold
    )
    # don't let weights go below 1e-6
    weights = np.clip(weights, 1e-6, None)

    # only continue on the columns with non-zero mean
    objNZ = obj[:, ~obj.var["allZero"]]
    weights = weights[:, ~obj.var["allZero"]]

    if alphaInit is None:
        # this rough dispersion estimate (alpha_hat)
        # is for estimating mu
        # and for the initial starting point for line search
        roughDisp = roughDispEstimate(y=objNZ.counts(normalized=True), x=modelMatrix)
        momentsDisp = momentsDispEstimate(objNZ)
        alpha_hat = np.minimum(roughDisp, momentsDisp)
    else:
        if np.isscalar(alphaInit):
            alpha_hat = np.repeat(alphaInit, objNZ.n_var)
        else:
            if len(alphaInit) != objNZ.n_var:
                raise ValueError("len(alphaInit) and objNZ.n_var mismatch")
            alpha_hat = alphaInit

    # bound the rough estimated alpha between minDisp and maxDisp for numeric stability
    maxDisp = np.maximum(10, obj.n_obs)
    alpha_init = np.clip(alpha_hat, minDisp, maxDisp)
    alpha_hat_new = alpha_init
    alpha_hat = alpha_init

    if niter <= 0:
        raise ValueError("niter should be strictly positive")

    # use a linear model to estimate the expected counts
    # if the number of groups according to the model matrix
    # is equal to the number of columns
    if linearMu is None:
        modelMatrixGroups = Factor(
            [tuple(modelMatrix[i]) for i in range(modelMatrix.shape[0])]
        )
        linearMu = modelMatrixGroups.nlevels() == modelMatrix.shape[1]
        # also check for weights (then can't do linear mu)
        if useWeights:
            linearMu = False

    # below, iterate between mean and dispersion estimation (niter) times
    fitidx = np.repeat(True, objNZ.n_vars)
    mu = np.zeros(objNZ.shape)
    dispIter = np.zeros(objNZ.n_vars)
    # bound the estimated count by 'minmu'
    # this helps make the fitting more robust
    # because 1/mu occurs in the weights for the NB GLM
    for iter in range(niter):
        if not linearMu:
            fitMu = fitNbinomGLMs(
                objNZ[:, fitidx],
                alpha_hat=alpha_hat[fitidx],
                modelMatrix=modelMatrix,
                type_=type_,
            )["mu"]
        else:
            fitMu = linearModelMuNormalized(objNZ[:, fitidx], modelMatrix)

        fitMu[fitMu < minmu] = minmu
        mu[:, fitidx] = fitMu

        # use of kappa_0 in backtracking search
        # inital proposal = log(alpha) + kappa_0 * deriv. of log lik. w.r.t. log(alpha)
        # use log(minDisp/10) to stop if dispersions going to -infinity
        if type_ == "DESeq2":
            dispRes = fitDispWrapper(
                y=objNZ.counts()[:, fitidx],
                x=modelMatrix,
                mu_hat=fitMu,
                log_alpha=np.log(alpha_hat)[fitidx],
                log_alpha_prior_mean=np.log(alpha_hat)[fitidx],
                log_alpha_prior_sigmasq=1,
                min_log_alpha=np.log(minDisp / 10),
                kappa_0=kappa_0,
                tol=dispTol,
                maxit=maxit,
                usePrior=False,
                weights=weights,
                useWeights=useWeights,
                weightThreshold=weightThreshold,
                useCR=useCR,
            )

            dispIter[fitidx] = dispRes["iter"]
            alpha_hat_new[fitidx] = np.minimum(np.exp(dispRes["log_alpha"]), maxDisp)
            last_lp = dispRes["last_lp"]
            initial_lp = dispRes["initial_lp"]
            # only rerun those cols which moved

        elif type_ == "glmGamPoi":
            raise NotImplementedError("glmGamPoi not implemented")

        fitidx = np.abs(np.log(alpha_hat_new) - np.log(alpha_hat)) > 0.5
        alpha_hat = alpha_hat_new
        if np.sum(fitidx) == 0:
            break

    # dont accept moves if the log posterior did not
    # increase by more than one millionth,
    # and set the small estimates to the minimum dispersion
    dispGeneEst = alpha_hat
    if niter == 1:
        noIncrease = last_lp < initial_lp + np.abs(initial_lp) / 1e6
        dispGeneEst[noIncrease] = alpha_init[noIncrease]
    # did not reach the maximum and iterated more than once
    dispGeneEstConv = dispIter < maxit & (dispIter > 1)

    # if lacking convergence from fitDisp() (C++)...
    refitDisp = ~dispGeneEstConv & (dispGeneEst > minDisp * 10)
    if np.sum(refitDisp) > 0:
        dispGrid = fitDispGridWrapper(
            y=objNZ.counts()[:, refitDisp],
            x=modelMatrix,
            mu=mu[:, refitDisp],
            log_alpha_prior_mean=np.repeat(0, np.sum(refitDisp)),
            log_alpha_prior_sigmasq=1,
            usePrior=False,
            weights=weights[:, refitDisp],
            useWeights=useWeights,
            weightThreshold=weightThreshold,
            useCR=useCR,
        )
        dispGeneEst[refitDisp] = dispGrid

    dispGeneEst = np.clip(dispGeneEst, minDisp, maxDisp)

    obj.var["dispGeneEst"] = buildVectorWithNACols(dispGeneEst, obj.var["allZero"])
    obj.var.type["dispGeneEst"] = "intermediate"
    obj.var.description["dispGeneEst"] = "gene-wise estimates of dispersion"
    obj.var["dispGeneIter"] = buildVectorWithNACols(dispIter, obj.var["allZero"])
    obj.var.type["dispGeneIter"] = "intermediate"
    obj.var.description["dispGeneIter"] = "number of iterations for gene-wise"

    obj.layers["mu"] = buildMatrixWithNACols(mu, obj.var["allZero"])

    return obj


def estimateDispersionsFit(obj, fitType="parametric", minDisp=1e-8, quiet=False):
    """
    Low-level function to fit dispersion estimates

    Normal users should instead use :meth:`.DESeqDataSet.estimateDispersions`.
    This low-level function is called by
    :meth:`.DESeqDataSet.estimateDispersions`, but is exported and documented
    for non-standard usage.  For instance, it is possible to replace fitted
    values with a custom fit and continue with the maximum a posteriori
    dispersion estimate.

    Arguments
    ---------
    obj : DESeqDataSet
        the input dataset
    fitType : "parametric", "local", "mean" or "glmGamPoi"
        the type of fitting of dispersions to the mean intensity.
        See :meth:`.DESeqDataSet.estimateDispersions` for description.
    minDisp : float
        small value for the minimum dispersion, to allow for calculations in
        log scale, one order of magnitude above this value is used as a test
        for inclusion in mean-dispersion fitting
    quiet : bool
        whether to print messages at each step

    Returns
    -------
    DESeqDataSet
        the input :code:`obj` with fitted dispersion estimates
    """
    if "allZero" not in obj.var:
        obj = obj.getBaseMeansAndVariances()

    objNZ = obj[:, ~obj.var["allZero"]]
    useForFit = objNZ.var["dispGeneEst"] > 100 * minDisp
    if useForFit.sum() == 0:
        raise ValueError(
            """
            all gene-wise dispersion estimates are within 2 orders of magnitude
            from the minimum value, and so the standard curve fitting techniques will not work.
            One can instead use the gene-wise estimates as final estimates:
            dds = estimateDispersionsGeneEst(dds)
            dds.dispersions = dds.var["dispGeneEst"]
            ... then continue with testing using nbinonWaldTest or nbinomLRT
            """
        )

    if fitType not in ["parametric", "local", "mean", "glmGamPoi"]:
        raise ValueError(f"incorrect value for fitType: {fitType}")

    if fitType == "parametric":
        try:
            dispFunction = parametricDispersionFit(
                objNZ.var["baseMean"][useForFit], objNZ.var["dispGeneEst"][useForFit]
            )
        except RuntimeError as e:
            logging.info(f"parametric fit failed with {e}")
            logging.info(
                "note: fitType='parametric', but the dispersion trend was not well captured by the function: y = a/x + b, and a local regression fit was automatically substituted. Specify fitType='local' or 'mean' to avoid this message next time."
            )
            fitType = "local"

    if fitType == "local":
        raise NotImplementedError()
        # dispFunction = localDispersionFit(
        # means = objNZ.var["baseMean"][useForFit],
        # disps = objNZ.var["dispGeneEst"][useForFit],
        # minDisp = minDisp)

    if fitType == "mean":
        useForMean = objNZ.var["dispGeneEst"] > 10 * minDisp
        useForMean = useForMean & ~np.isnan(objNZ.var["dispGeneEst"])
        meanDisp = trim_mean(objNZ.var["dispGeneEst"][useForMean], 0.001)
        dispFunction = lambda means: meanDisp
        dispFunction.mean = meanDisp

    if fitType == "glmGamPoi":
        raise NotImplementedError()

    # store the dispersion function and attributes
    dispFunction.fitType = fitType
    obj.setDispFunction(dispFunction)

    return obj


def estimateDispersionsMAP(
    obj,
    outlierSD=2,
    dispPriorVar=None,
    minDisp=1e-8,
    kappa_0=1,
    dispTol=1e-6,
    maxit=100,
    useCR=True,
    weightThreshold=1e-2,
    modelMatrix=None,
    type_="DESeq2",
    quiet=False,
):
    """
    Low-level function to fit dispersion estimates

    Normal users should instead use :meth:`.DESeqDataSet.estimateDispersions`.
    This low-level function is called by
    :meth:`.DESeqDataSet.estimateDispersions`, but is exported and documented
    for non-standard usage.  For instance, it is possible to replace fitted
    values with a custom fit and continue with the maximum a posteriori
    dispersion estimate.

    :func:`estimateDispersionsPriorVar` is called inside this function, and
    stores the dispersion prior variance as an attribute of
    :attr:`.DESeqDataSet.dispersionFunction`, which can be manually provided to
    :code:`estimateDispersionsMAP` for parallel execution.

    Arguments
    ---------
    obj : DESeqDataSet
        the input dataset
    outlierSD : int
        the number of standard deviations of log gene-wise estimates above the
        prior mean (fitted value), above which dispersion estimates will be
        labelled outliers. Outliers will keep their original value and not be
        shrunk using the prior.
    dispPriorVar : array-like
        the variance of the normal priori on the log dispersions. If not
        supplied, it is calculated as the difference between the mean squared
        residuals of gene-wise estimates to the fitted dispersion and the
        expected sampling variance of the log dispersion.
    minDisp : float
        small value for the minimum dispersion, to allow for calculations in
        log scale, one order of magnitude above this value is used as a test
        for inclusion in mean-dispersion fitting
    kappa_0 : float
        parameter used in setting the initial proposal in backtracking search,
        higher :code:`kappa_0` results in larger steps
    dispTol : float
        parameter to test for convergence of log dispersion, stop when increase
        in log posterior is less than :code:`dispTol`
    maxit : int
        maximum number of iterations to allow for convergence
    useCR : bool
        whether to use Cox-Reid adjustment
    weightThreshold : float
        threshold for subsetting the design matrix and GLM weights for
        calculating the Cox-Reid correction
    quiet : bool
        whether to print messages at each step
    modelMatrix : array-like
        for advanced use only, a substitute model matrix for gene-wise and MAP
        dispersion estimation
    type_ : "DESeq2" or "glmGamPoi"
        specify if the glmGamPoi package is used to calculate the dispersion.
        This can be significantly faster if there are many replicates with
        small counts.

    Returns
    -------
    DESeqDataSet
        the input :code:`obj` with final MAP dispersion estimates
    """
    if type_ not in ["DESeq2", "glmGamPoi"]:
        raise ValueError(f"invalid value for type_: {type_}")

    if "allZero" not in obj.var:
        obj = obj.getBaseMeansAndVariances()
    if "dispersion" in obj.var:
        logging.info("found already estimated dispersions, removing these")
        del obj.var["dispersion"]
        del obj.var["dispOutlier"]
        del obj.var["dispMAP"]
        del obj.var["dispIter"]
        del obj.var["dispConv"]

    if modelMatrix is None:
        modelMatrix = obj.design

    # fill in the calculated dispersion prior variance
    if dispPriorVar is None:
        # if no gene-wise estimates above minimum
        if np.nansum(obj.var["dispGeneEst"] >= 100 * minDisp) == 0:
            logging.warnings.warn(
                f"all genes have dispersion estimates < {100*minDisp}, returning disp = {10*minDisp}"
            )
            obj.var["dispersion"] = buildVectorWithNACols(
                10 * minDisp, obj.var["allZero"]
            )
            obj.var.type["dispersion"] = "intermediate"
            obj.var.description["dispersion"] = "final estimates of dispersion"
            dispFn = obj.dispersionFunction
            dispFn.dispPriorVar = 0.25
            obj.setDispFunction(dispFn, estimateVar=False)
            return obj

        dispPriorVar = estimateDispersionsPriorVar(obj, modelMatrix=modelMatrix)
        dispFn = obj.dispersionFunction
        dispFn.dispPriorVar = dispPriorVar
        obj.setDispFunction(dispFn, estimateVar=False)

    # use weights if they are present in obj.layers
    (obj, weights, useWeights) = getAndCheckWeights(
        obj, modelMatrix, weightThreshold=weightThreshold
    )

    objNZ = obj[:, ~obj.var["allZero"]]
    weights = weights[:, ~obj.var["allZero"]]
    varLogDispEsts = obj.dispersionFunction.varLogDispEsts

    # set prior variance for fitting dispersion
    log_alpha_prior_sigmasq = dispPriorVar

    # get previously calculated mu
    mu = objNZ.layers["mu"]

    if type_ == "DESeq2":
        # start fitting at gene estimate unless the points are one order of magnitude
        # below the fitted line, then start at fitted line
        dispInit = np.where(
            objNZ.var["dispGeneEst"] > 0.1 * objNZ.var["dispFit"],
            objNZ.var["dispGeneEst"],
            objNZ.var["dispFit"],
        )

        # if any missing values, fill in the fitted values to initialize
        dispInit[np.isnan(dispInit)] = objNZ.var["dispFit"][np.isnan(dispInit)]

        # run with prior
        dispResMAP = fitDispWrapper(
            y=objNZ.counts(),
            x=modelMatrix,
            mu_hat=mu,
            log_alpha=np.log(dispInit),
            log_alpha_prior_mean=np.log(objNZ.var["dispFit"]),
            log_alpha_prior_sigmasq=log_alpha_prior_sigmasq,
            min_log_alpha=np.log(minDisp / 10),
            kappa_0=kappa_0,
            tol=dispTol,
            maxit=maxit,
            usePrior=True,
            weights=weights,
            useWeights=useWeights,
            weightThreshold=weightThreshold,
            useCR=useCR,
        )

        # prepare dispersions for storage
        dispMAP = np.exp(dispResMAP["log_alpha"])
        dispIter = dispResMAP["iter"]

        # when lacking convergence from fitDisp()
        # we use a function to maximize dispersion parameter
        # along an adaptive grid
        dispConv = dispResMAP["iter"] < maxit
        refitDisp = ~dispConv
        if np.sum(refitDisp) > 0:
            dispGrid = fitDispGridWrapper(
                y=objNZ.counts()[:, refitDisp],
                x=modelMatrix,
                mu=mu[:, refitDisp],
                log_alpha_prior_mean=np.log(objNZ.var["dispFit"])[refitDisp],
                log_alpha_prior_sigmasq=log_alpha_prior_sigmasq,
                usePrior=True,
                weights=weights[:, refitDisp],
                useWeights=useWeights,
                weightThreshold=weightThreshold,
                useCR=True,
            )
            dispMAP[refitDisp] = dispGrid
    elif type_ == "glmGamPoi":
        raise NotImplementedError()

    # bound the dispersion estimate between minDisp and maxDisp for numeric stability
    maxDisp = np.maximum(10, obj.n_obs)
    dispMAP = np.clip(dispMAP, minDisp, maxDisp)

    dispersionFinal = dispMAP.copy()

    # detect outliers which have gene-wise estimates
    # outlierSD * standard deviation of log gene-wise estimates
    # above the fitted mean (prior mean)
    # and keep the original gene-est value for these.
    # Note: we use the variance of log dispersions estimates
    # from all the genes, not only those from below
    dispOutlier = np.log(objNZ.var["dispGeneEst"]) > np.log(
        objNZ.var["dispFit"]
    ) + outlierSD * np.sqrt(varLogDispEsts)
    dispOutlier[np.isnan(dispOutlier)] = False
    dispersionFinal[dispOutlier] = objNZ.var["dispGeneEst"][dispOutlier]

    obj.var["dispersion"] = buildVectorWithNACols(dispersionFinal, obj.var["allZero"])
    obj.var.type["dispersion"] = "intermediate"
    obj.var.description["dispersion"] = "final estimates of dispersion"
    obj.var["dispIter"] = buildVectorWithNACols(dispIter, obj.var["allZero"])
    obj.var.type["dispIter"] = "intermediate"
    obj.var.description["dispIter"] = "number of iterations"
    obj.var["dispOutlier"] = buildVectorWithNACols(dispOutlier, obj.var["allZero"])
    obj.var.type["dispOutlier"] = "intermediate"
    obj.var.description["dispOutlier"] = "dispersion flagged as outlier"
    obj.var["dispMAP"] = buildVectorWithNACols(dispMAP, obj.var["allZero"])
    obj.var.type["dispMAP"] = "intermediate"
    obj.var.description["dispMAP"] = "maximum a posteriori estimate"

    return obj


def estimateDispersionsPriorVar(obj, minDisp=1e-8, modelMatrix=None):
    """
    Low-level function to fit dispersion estimates

    Normal users should instead use :meth:`.DESeqDataSet.estimateDispersions`.
    This low-level function is called by
    :meth:`.DESeqDataSet.estimateDispersions`, but is exported and documented
    for non-standard usage.  For instance, it is possible to replace fitted
    values with a custom fit and continue with the maximum a posteriori
    dispersion estimate.

    This function is called inside :func:`estimateDispersionsMAP`, and stores
    the dispersion prior variance as an attribute of
    :attr:`.DESeqDataSet.dispersionFunction`, which can be manually provided to
    :func:`estimateDispersionsMAP` for parallel execution.

    Arguments
    ---------
    obj : DESeqDataSet
        the input dataset
    minDisp : float
        small value for the minimum dispersion, to allow for calculations in
        log scale, one order of magnitude above this value is used as a test
        for inclusion in mean-dispersion fitting
    modelMatrix : array-like
        for advanced use only, a substitute model matrix for gene-wise and MAP
        dispersion estimation

    Returns
    -------
    DESeqDataSet
        the input :code:`obj` with final MAP dispersion estimates
    """
    objNZ = obj[:, ~obj.var["allZero"]]
    aboveMinDisp = objNZ.var["dispGeneEst"] >= 100 * minDisp
    if modelMatrix is None:
        modelMatrix = objNZ.design
    # estimate the variance of the distribution of the
    # log dispersion estimates around the fitted value
    dispResiduals = np.log(objNZ.var["dispGeneEst"]) - np.log(objNZ.var["dispFit"])
    if np.nansum(aboveMinDisp) == 0:
        raise ValueError("no data found which is greater than minDisp")

    varLogDispEsts = obj.dispersionFunction.varLogDispEsts

    m, p = modelMatrix.shape

    # if the residual degrees of freedom is between 1 and 3, the distribution
    # of log dispersions is especially asymmetric and poorly estimated
    # by the MAD. We then use an alternate estimator, a Monte Carlo
    # approach to match the distribution.
    if (m - p) <= 3 and m > p:
        # TODO implement Monte Carlo re-estimation
        pass

    # estimate the expected sampling variance of the log estimates
    # Var(log(cX)) = Var(log(X))
    # X ~ chi-squared with m - p degrees of freedom
    if m > p:
        expVarLogDisp = polygamma(1, (m - p) / 2)
        # set the variance of the prior using these two estimates
        # with a minimum of 0.25
        dispPriorVar = np.maximum(varLogDispEsts - expVarLogDisp, 0.25)
    else:
        # we have m = p, so do not try to subtract sampling variance
        dispPriorVar = varLogDispEsts
        expVarLogDisp = 0

    return dispPriorVar


def checkForExperimentalReplicates(obj, modelMatrix):
    if modelMatrix is None:
        modelMatrix = obj.design

    if modelMatrix.shape[0] == modelMatrix.shape[1]:
        raise ValueError(
            """
                The design matrix has the same number of samples and coefficients to fit,
                so estimation of dispersion is not possible. Treating samples
                as replicates is not supported."""
        )


def roughDispEstimate(y, x):
    """rough dispersion estimate using counts and fitted values

    Arguments
    ---------
    y : array-like
        normalized counts matrix (shape nobs x nvar)
    x : array-like
        design matrix (shape nobs x nd)
    """
    # must be positive
    mu = linearModelMu(y, x)
    mu = np.clip(mu, 1, None)

    m, p = x.shape

    # an alternate rough estimator with higher mean squared or absolute error
    # (colSums( (y - mu)^2/(mu * (m - p)) ) - 1) / colMeans(mu)

    # rough disp estimates will be adjusted up to minDisp later
    est = np.sum(((y - mu) ** 2 - mu) / mu**2, 0) / (m - p)
    return np.clip(est, 0, None)


def linearModelMu(y, x):
    """
    Arguments
    ---------
    y : array-like
        counts matrix
    x : array-like
        design matrix (as many rows as y)
    """
    # NB: in the R version, y is transposed compared to this Python version
    # original R version: ((x Rinv Q.T) y.T).T
    # optimized R version: (y Q) (x Rinv).T
    # we choose to try (x Rinv) (Q.T y), which seems to minimize the size of intermediate matrices (assuming p <= nobs)

    # NB: the R code, and the code below, assumes that p <= nobs.
    # this is guaranteed by the way we check that the design matrix is full rank.
    (Q, R) = np.linalg.qr(x)
    Rinv = np.linalg.solve(R, np.identity(R.shape[0]))
    return (x @ Rinv) @ (Q.T @ y)


def linearModelMuNormalized(obj, x):
    """
    Arguments
    ---------
    y : DESeqDataSet
    x : array-like
        design matrix
    """
    norm_cts = obj.counts(normalized=True)
    muhat = linearModelMu(norm_cts, x)
    nf = obj.getSizeOrNormFactors()
    return muhat * nf


def momentsDispEstimate(obj):
    if obj.normalizationFactors is not None:
        xim = np.mean(1 / np.mean(obj.normalizationFactors, 1))
    else:
        xim = np.mean(1 / obj.sizeFactors)

    bv = obj.var["baseVar"]
    bm = obj.var["baseMean"]
    return (bv - xim * bm) / bm**2


def parametricDispersionFit(means, disps):
    """Estimate a parametric fit of dispersion to the mean intensity

    Arguments
    ---------
    means : ??
    disps : ??
    """
    logging.warnings.simplefilter("ignore", DomainWarning)
    coefs = [0.1, 1]
    iter_ = 0
    while True:
        residuals = disps / (coefs[0] + coefs[1] / means)
        good = (residuals > 1e-4) & (residuals < 15)
        # check for glm convergence below to exit while loop
        glm_gamma = sm.GLM(
            disps[good],
            sm.add_constant(1 / means[good]),
            family=sm.families.Gamma(link=sm.families.links.identity()),
        )
        fit = glm_gamma.fit(start_params=coefs)
        oldcoefs = coefs.copy()
        coefs = fit.params

        if not np.all(coefs > 0):
            raise RuntimeError("parametric dispersion fit failed")
        if np.sum(np.log(coefs / oldcoefs) ** 2) < 1e-6 and fit.converged:
            break
        iter_ = iter_ + 1
        if iter_ > 10:
            raise RuntimeError("dispersion fit did not converge")

    coefs.index = ["asymptDisp", "extraPois"]
    ans = lambda q: coefs[0] + coefs[1] / q
    ans.coefficients = coefs
    return ans



import unittest
import numpy as np

from inmoose.deseq2 import makeExampleDESeqDataSet, DESeq, nbinomWaldTest


class Test(unittest.TestCase):
    def test_QR(self):
        """test that not using QR works as expected"""

        dds = makeExampleDESeqDataSet(n=100, betaSD=1, seed=42)
        dds = DESeq(dds, quiet=True)
        ddsNoQR = nbinomWaldTest(dds.copy(), useQR=False)
        res = dds.results()
        resNoQR = ddsNoQR.results()
        self.assertTrue(np.allclose(res.log2FoldChange, resNoQR.log2FoldChange))



# -----------------------------------------------------------------------------
# Copyright (C) 2019-2020 Yuqing Zhang
# Copyright (C) 2022-2023 Maximilien Colange

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
# -----------------------------------------------------------------------------

# This file is based on the file 'R/helper_seq.R' of the Bioconductor sva package (version 3.44.0).


import numpy as np
from scipy.stats import nbinom


def vec2mat(vec, n_times):
    """
    Expand a vector into matrix (columns as the original vector)
    """
    vec = np.asarray(vec)
    vec = vec.reshape(vec.shape[0], 1)
    return np.full((vec.shape[0], n_times), vec)


def match_quantiles(counts_sub, old_mu, old_phi, new_mu, new_phi):
    """
    Match quantiles from a source negative binomial distribution to a target
    negative binomial distribution.

    Arguments
    ---------
    counts_sub : array_like
        the original data following the source distribution
    old_mu : array_like
        the mean of the source distribution
    old_phi : array_like
        the dispersion of the source distribution
    new_mu : array_like
        the mean of the target distribution
    new_phi : array_like
        the dispersion of the target distribution

    Returns
    -------
    ndarray
        adjusted data, corresponding in the target distribution to the same
        quantiles as the input data in the source distribution
    """
    new_counts_sub = np.full(counts_sub.shape, np.nan)

    i = counts_sub <= 1
    new_counts_sub[i] = counts_sub[i]
    i = np.logical_not(i)
    old_size = np.full(old_mu.shape, 1 / old_phi.reshape(old_mu.shape[0], 1))
    new_size = np.full(new_mu.shape, 1 / new_phi.reshape(new_mu.shape[0], 1))
    old_prob = old_size / (old_size + old_mu)
    new_prob = new_size / (new_size + new_mu)
    tmp_p = nbinom.cdf(counts_sub[i] - 1, old_size[i], old_prob[i])
    new_counts_sub[i] = np.where(
        np.abs(tmp_p - 1) < 1e-4,
        counts_sub[i],
        1 + nbinom.ppf(tmp_p, new_size[i], new_prob[i]),
    )

    # Original (pythonized) R code for reference
    #
    # for a in range(counts_sub.shape[0]):
    #    for b in range(counts_sub.shape[1]):
    #        if counts_sub[a,b] <= 1:
    #            new_counts_sub[a,b] = counts_sub[a,b]
    #        else:
    #            tmp_p = pnbinom_opt(counts_sub[a,b]-1, mu=old_mu[a,b], size=1/old_phi[a])
    #            if abs(tmp_p-1) < 1e-4:
    #                # for outlier count, if tmp_p==1, qnbinom(tmp_p) will return Inf values -> use original count instead
    #                new_counts_sub[a,b] = counts_sub[a,b]
    #            else:
    #                new_counts_sub[a,b] = 1+qnbinom_opt(tmp_p, mu=new_mu[a,b], size=1/new_phi[a])

    return new_counts_sub



import unittest
import numpy as np
import pandas as pd

from inmoose.pycombat.covariates import make_design_matrix
from inmoose.pycombat.pycombat_norm import check_mean_only
from inmoose.pycombat.pycombat_norm import (
    compute_prior,
    postmean,
    postvar,
    it_sol,
    int_eprior,
)
from inmoose.pycombat.pycombat_norm import calculate_mean_var, calculate_stand_mean
from inmoose.pycombat.pycombat_norm import standardise_data, fit_model, adjust_data
from inmoose.pycombat import pycombat_norm


class test_pycombat(unittest.TestCase):
    @classmethod
    def setUpClass(self):
        matrix = np.transpose(
            [
                np.random.normal(size=1000, loc=3, scale=1),
                np.random.normal(size=1000, loc=3, scale=1),
                np.random.normal(size=1000, loc=3, scale=1),
                np.random.normal(size=1000, loc=2, scale=0.6),
                np.random.normal(size=1000, loc=2, scale=0.6),
                np.random.normal(size=1000, loc=4, scale=1),
                np.random.normal(size=1000, loc=4, scale=1),
                np.random.normal(size=1000, loc=4, scale=1),
                np.random.normal(size=1000, loc=4, scale=1),
            ]
        )

        self.matrix = pd.DataFrame(
            data=matrix,
            columns=["sample_" + str(i + 1) for i in range(9)],
            index=["gene_" + str(i + 1) for i in range(1000)],
        )

        # test normal execution
        self.batch = np.asarray([1, 1, 1, 2, 2, 3, 3, 3, 3])
        self.matrix_adjusted = pycombat_norm(self.matrix, self.batch)

        # useful constants for unit testing
        ref_batch = None
        mean_only = False
        par_prior = False
        precision = None
        mod = None
        self.dat = self.matrix.values
        vci = make_design_matrix(
            self.dat, self.batch, covar_mod=mod, ref_batch=ref_batch
        )
        self.dat = vci.counts
        self.design = np.transpose(vci.design)
        self.batchmod = vci.batch_mod
        self.batches = [vci.batch_composition[b] for b in vci.batch.categories]
        self.n_batches = [len(v) for v in self.batches]
        self.batch = vci.batch
        ref = vci.ref_batch_idx
        self.n_batch = vci.n_batch
        self.n_array = self.dat.shape[1]

        self.B_hat, self.grand_mean, self.var_pooled = calculate_mean_var(
            self.design,
            self.batches,
            ref,
            self.dat,
            self.n_batches,
            self.n_batch,
            self.n_array,
        )
        self.stand_mean = calculate_stand_mean(
            self.grand_mean, self.n_array, self.design, self.n_batch, self.B_hat
        )
        self.s_data = standardise_data(
            self.dat, self.stand_mean, self.var_pooled, self.n_array
        )
        self.gamma_star, self.delta_star, self.batch_design = fit_model(
            self.design,
            self.n_batch,
            self.s_data,
            self.batches,
            mean_only,
            par_prior,
            precision,
            ref,
        )
        self.bayes_data = adjust_data(
            self.s_data,
            self.gamma_star,
            self.delta_star,
            self.batch_design,
            self.n_batches,
            self.var_pooled,
            self.stand_mean,
            self.n_array,
            ref,
            self.batches,
            self.dat,
        )

    def test_compute_prior(self):
        print("aprior", compute_prior("a", self.gamma_star, False))
        self.assertEqual(compute_prior("a", self.gamma_star, True), 1)
        print("bprior", compute_prior("b", self.gamma_star, False))
        self.assertEqual(compute_prior("b", self.gamma_star, True), 1)

    def test_postmean(self):
        self.assertEqual(
            np.shape(
                postmean(
                    self.gamma_star, self.delta_star, self.gamma_star, self.delta_star
                )
            ),
            np.shape(self.gamma_star),
        )

    def test_postvar(self):
        self.assertEqual(np.sum(postvar([2, 4, 6], 2, 1, 1) - [2, 3, 4]), 0)

    # def test_it_sol(self):
    #    ()

    # def test_int_eprior(self):
    #    ()

    @unittest.skip("automate the verification")
    def test_check_mean_only(self):
        check_mean_only(True)
        check_mean_only(False)
        print("Only one line of text should have been printed above.")

    def test_calculate_mean_var(self):
        self.assertEqual(np.shape(self.B_hat)[0], np.shape(self.design)[0])
        self.assertEqual(np.shape(self.grand_mean)[0], np.shape(self.dat)[0])
        self.assertEqual(np.shape(self.var_pooled)[0], np.shape(self.dat)[0])

    def test_calculate_stand_mean(self):
        self.assertEqual(np.shape(self.stand_mean), np.shape(self.dat))

    def test_standardise_data(self):
        self.assertEqual(np.shape(self.s_data), np.shape(self.dat))

    def test_fit_model(self):
        self.assertEqual(np.shape(self.gamma_star)[1], np.shape(self.dat)[0])
        self.assertEqual(np.shape(self.delta_star)[1], np.shape(self.dat)[0])
        self.assertEqual(np.shape(self.batch_design), np.shape(self.design))

    def test_adjust_data(self):
        self.assertEqual(np.shape(self.bayes_data), np.shape(self.dat))

    def test_pycombat(self):
        self.assertEqual(np.shape(self.matrix), np.shape(self.matrix_adjusted))
        self.assertTrue(
            np.mean(self.matrix.values) - np.mean(self.matrix_adjusted.values)
            <= np.mean(self.matrix.values) * 0.05
        )
        self.assertTrue(
            np.var(self.matrix_adjusted.values) <= np.var(self.matrix.values)
        )
        # test raise error for single sample batch
        with self.assertRaisesRegex(
            ValueError,
            r"Batches 4 contain a single sample, which is not supported for batch effect correction. Please review your inputs.",
        ):
            pycombat_norm(self.matrix, np.asarray([1, 1, 1, 2, 2, 3, 3, 3, 4]))



# -----------------------------------------------------------------------------
# Copyright (C) 2022-2023 M. Colange

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
# -----------------------------------------------------------------------------

from scipy.stats import nbinom, norm, t

from .stats_cpp import nbinom_logpmf


def rnbinom(n, size, mu, seed=None):
    """mimic R rnbinom function, to draw samples from a Negative Binomial distribution.

    The (:math:`size`, :math:`p`) parameterization used in R is the same as in scipy.stats:
    :math:`p = 1 / (1 + \mu/size) = size / (size + \mu)`.

    Arguments
    ---------
    n : int or tuple of ints
        shape of the output. If n = (n1, n2, ..., np) then n1*n2*...*np random samples are drawn.
    size : float or array-like
        size parameter of the Negative Binomial distribution.
        all values must be positive
    mu : float or array-like
        mean parameter of the Negative Binomial distribution
        all values must be positive
    seed : int, optional
        pass a seed to the underlying RNG. If `None`, then the RNG is seeded using unpredictable entropy from the system.
        See the documentation of scipy.stats about RNG seeding for more details.
    """
    p = size / (size + mu)
    return nbinom(size, p).rvs(n, random_state=seed)


def dnbinom_mu(x, size, mu, log=False):
    """mimic R dnbinom_mu function, to compute the density function of a Negative Binomial distribution.

    The (size, prob) parameterization used in R dnbinom is the same as in scipy.stats:
        mu = size * (1 - prob) / prob = size * (1/prob - 1)
    so
        mu / size + 1 = 1/prob
    hence
        prob = 1 / (1 + mu/size) = size / (size + mu)

    Arguments
    ---------
    x : array-like
        points where the density function is evaluated
    size : float or array-like
        size parameter of the Negative Binomial distribution.
        all values must be positive
    mu : float or array-like
        mean parameter of the Negative Binomial distribution
        all values must be positive
    log: bool, optional
        switch to output the density in normal (log=False) or log (log=True) scale
    """
    p = size / (size + mu)
    if log:
        return nbinom_logpmf(x, size, p)
    else:
        return nbinom.pmf(x, n=size, p=p)


def dnorm(x, mean, sd, log=False):
    """mimic R dnorm function, to compute the density function of a normal distribution.

    Arguments
    ---------
    x : array-like
        points where the density function is evaluated
    mean : float or array-like
        mean of the normal distribution
    sd : float or array-like
        standard deviation of the normal distribution
    log : bool, optional
        switch to output the density in normal (log=False) or log (log=True) scale
    """
    if log:
        return norm.logpdf(x, loc=mean, scale=sd)
    else:
        return norm.pdf(x, loc=mean, scale=sd)


def pnorm(q, mean=0, sd=1, lower_tail=True, log_p=False):
    """mimic R pnorm function, to compute the distribution function of a normal distribution.

    Arguments
    ---------
    q : array-like
        points where the distribution function is evaluated
    mean : float or array-like
        mean of the normal distribution
    sd : float or array-like
        standard deviation of the normal distribution
    log: bool, optional
        switch to output the distribution in normal (log=False) or log (log=True) scale
    lower_tail : bool, optional
        switch to output P[X <= q] (lower_tail=True) or P[X > q] (lower_tail=False)
    """
    if lower_tail and log_p:
        f = norm.logcdf
    elif lower_tail and not log_p:
        f = norm.cdf
    elif not lower_tail and log_p:
        f = norm.logsf
    elif not lower_tail and not log_p:
        f = norm.sf

    return f(q, loc=mean, scale=sd)


def pt(q, df, lower_tail=True, log_p=False):
    """mimic R pt function, to compute the distribution function of a Student t distribution.

    Arguments
    ---------
    q : array-like
        points where the distribution function is evaluated
    df : int or array-like
        number of degrees of freedom of the Student t distribution
    log: bool, optional
        switch to output the distribution in normal (log=False) or log (log=True) scale
    lower_tail : bool, optional
        switch to output P[X <= q] (lower_tail=True) or P[X > q] (lower_tail=False)
    """
    if lower_tail and log_p:
        f = t.logcdf
    elif lower_tail and not log_p:
        f = t.cdf
    elif not lower_tail and log_p:
        f = t.logsf
    elif not lower_tail and not log_p:
        f = t.sf

    return f(q, df=df)



# -----------------------------------------------------------------------------
# Copyright (C) 2008-2022 Yunshun Chen, Aaron TL Lun, Davis J McCarthy, Matthew E Ritchie, Belinda Phipson, Yifang Hu, Xiaobei Zhou, Mark D Robinson, Gordon K Smyth
# Copyright (C) 2022-2023 Maximilien Colange

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
# -----------------------------------------------------------------------------

# This file is based on the file 'R/mglmLevenberg.R' of the Bioconductor edgeR package (version 3.38.4).


import numpy as np

from .edgepy_cpp import cxx_get_levenberg_start, cxx_fit_levenberg
from .makeCompressedMatrix import (
    _compressDispersions,
    _compressOffsets,
    _compressWeights,
)
from .utils import _isAllZero


def mglmLevenberg(
    y,
    design,
    dispersion=0,
    offset=0,
    weights=None,
    coef_start=None,
    start_method="null",
    maxit=200,
    tol=1e-6,
):
    """
    Fit genewise negative binomial GLMs with log-link using Levenberg damping to
    ensure convergence.

    This is a low-level work-horse used by higher-level functions, especially
    :func:`glmFit`. Most users will not need to call this function directly.

    This function fits a negative binomial GLM to each row of :code:`y`. The
    row-wise GLMs all have the same design matrix but possibly different
    dispersions, offsets and weights. It is low-level in that it
    operates on atomic objects (matrices and vectors).

    This function fits an arbitrary log-linear model to each response vector.
    It implements a Levenberg-Marquardt modification of the GLM scoring
    algorithm to prevent divergence. The main computation is implemented in C++.
    It treats the dispersion parameter of the negative binomial distribution
    as a known input.

    Arguments
    ---------
    y : array_like
        matrix of negative binomial counts. Rows for genes and columns for
        libraries.
    design : array_like
        design matrix of the GLM. Assumed to be full column rank
    dispersion : float or array_like
        scalar or vector giving the dispersion parameter for each GLM. Can be a
        scalar giving one value for all genes, or a vector of length equal to
        the number of genes giving genewise dispersions.
    offset : array_like
        vector or matrix giving the offset that is to be included in the log
        linear model predictor. Can be a scalar, a vector of length equal to the
        number of libraries, or a matrix of the same shape as :code:`y`.
    weights : matrix, optional
        vector or matrix of non-negative quantitative weights. Can be a vector
        of length equal to the number of libraries, or a matrix of the same
        shape as :code:`y`.
    coef_start : array_like, optional
        matrix of starting values for the linear model coefficient. Number of
        rows should agree with :code:`y` and number of columns should agree with
        :code:`design`. This argument does not usually need to be set as the
        automatic starting values perform well.
    start_method : str
        method used to generate starting values when :code:`coef_start = None`.
        Possible values are "null" to start from the null model of equal
        expression levels or :code:`y` to use the data as starting value for the
        mean.
    maxit : int
        the maximum number of iterations for the Fisher scoring algorithm. The
        iteration will be stopped when this limit is reached even if the
        convergence criterion has not been satisfied.
    tol : float
        the convergence tolerance.

    Returns
    -------
    tuple
        tuple wit the following components:

        - matrix of estimated coefficients for the linear models
        - matrix of fitted values
        - vector of residual deviances
        - number of iterations used
        - Boolean vector indicating genes for which the maximum damping was
          exceeded before convergence was achieved
    """
    # Check arguments
    y = np.asarray(y, order="F")
    (ngenes, nlibs) = y.shape
    if nlibs == 0 or ngenes == 0:
        raise ValueError("no data")

    # Check for negative, NA or non-finite values in the count matrix
    _isAllZero(y)

    # Check the design matrix
    design = np.asarray(design, order="F", dtype="double")
    # TODO check that all entries in design matrix are finite

    # Check dispersions, offsets, and weights
    offset = _compressOffsets(y, offset=offset)
    dispersion = _compressDispersions(y, dispersion)
    weights = _compressWeights(y, weights)

    # Initialize values for the coefficients at reasonable best guess with linear models
    if coef_start is None:
        if start_method not in ["null", "y"]:
            raise ValueError(f"invalid start_method {start_method}")
        beta = cxx_get_levenberg_start(
            y, offset, dispersion, weights, design, start_method == "null"
        )
    else:
        beta = np.asarray(coef_start, order="F", dtype="double")

    assert beta.shape == (y.shape[0], design.shape[1])
    # Check the arguments and call the C++ method
    output = cxx_fit_levenberg(y, offset, dispersion, weights, design, beta, tol, maxit)

    # Name the output and return it
    return output



# -----------------------------------------------------------------------------
# Copyright (C) 2008-2022 Yunshun Chen, Aaron TL Lun, Davis J McCarthy, Matthew E Ritchie, Belinda Phipson, Yifang Hu, Xiaobei Zhou, Mark D Robinson, Gordon K Smyth
# Copyright (C) 2022-2023 Maximilien Colange

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
# -----------------------------------------------------------------------------

# This file is based on the file 'R/mglmOneGroup.R' of the Bioconductor edgeR package (version 3.38.4).


import logging
import numpy as np

from .makeCompressedMatrix import (
    _compressDispersions,
    _compressOffsets,
    _compressWeights,
)
from .utils import _isAllZero
from .edgepy_cpp import cxx_fit_one_group


def mglmOneGroup(
    y,
    dispersion=0,
    offset=0,
    weights=None,
    coef_start=None,
    maxit=50,
    tol=1e-10,
    verbose=False,
):
    """
    Fit single-group negative-binomial GLMs genewise.

    This is a low-level work-horse used by higher-level functions, especially
    :func:`glmFit`. Most users will not need to call this function directly.

    This function fits a negative binomial GLM to each row of :code:`y`. The
    row-wise GLMs all have the same design matrix but possibly different
    dispersions, offsets and weights. It is low-level in that it
    operates on atomic objects (matrices and vectors).

    This function fits an intercept only moel to each response vector. In other
    words, it treats all the libraries as belonging to one group. It implements
    Fisher scoring with a score-statistic stopping criterion for each gene.  It
    treats the dispersion parameter of the negative binomial distribution as a
    known input.  Excellent starting values are available for the null model so
    this function seldom has any problems with convergence. It is used by other
    functions to compute the overall abundance for each gene.

    Arguments
    ---------
    y : array_like
        matrix of negative binomial counts. Rows for genes and columns for
        libraries.
    dispersion : float or array_like
        scalar or vector giving the dispersion parameter for each GLM. Can be a
        scalar giving one value for all genes, or a vector of length equal to
        the number of genes giving genewise dispersions.
    offset : array_like
        vector or matrix giving the offset that is to be included in the log
        linear model predictor. Can be a scalar, a vector of length equal to the
        number of libraries, or a matrix of the same shape as :code:`y`.
    weights : matrix, optional
        vector or matrix of non-negative quantitative weights. Can be a vector
        of length equal to the number of libraries, or a matrix of the same
        shape as :code:`y`.
    coef_start : array_like, optional
        matrix of starting values for the linear model coefficient. Number of
        rows should agree with :code:`y` and a single column. This argument does
        not usually need to be set as the automatic starting values perform well.
    maxit : int
        the maximum number of iterations for the Fisher scoring algorithm. The
        iteration will be stopped when this limit is reached even if the
        convergence criterion has not been satisfied.
    tol : float
        the convergence tolerance. Convergence if judged successful when the
        step size falls below :code:`tol` in absolute size.

    Returns
    -------
    ndarray
        vector of coefficients
    """
    # Check y
    # TODO check that y is a matrix and numeric
    _isAllZero(y)

    # Check dispersion
    dispersion = _compressDispersions(y, dispersion)

    # Check offset
    offset = _compressOffsets(y, offset=offset)

    # Check starting values
    if coef_start is None:
        coef_start = np.NaN
    coef_start = np.full((y.shape[0],), coef_start, dtype="double", order="F")

    # Check weights
    weights = _compressWeights(y, weights)

    # Fisher scoring iteration
    output = cxx_fit_one_group(y, offset, dispersion, weights, maxit, tol, coef_start)

    # Convergence achieved for all tags?
    if np.count_nonzero(output[1]) > 0:
        logging.debug(f"max iterations exceeded for {np.count_nonzero(output[1])} tags")

    return output[0]



import unittest
import numpy as np
import pandas as pd
import patsy
from scipy.optimize import minimize
import scipy.stats

from inmoose.utils import Factor, dnbinom_mu, dnorm
from inmoose.deseq2 import (
    DESeqDataSet,
    makeExampleDESeqDataSet,
    DESeq,
    estimateDispersionsGeneEst,
    estimateDispersionsFit,
    estimateDispersionsMAP,
)
from inmoose.deseq2.fitNbinomGLMs import fitNbinomGLMs
from inmoose.deseq2.deseq2_cpp import fitDisp


class Test(unittest.TestCase):
    def test_dispersion_errors(self):
        """test that expected errors are thrown during dispersion estimation"""
        dds = makeExampleDESeqDataSet(n=100, m=2)
        dds = dds.estimateSizeFactors()
        with self.assertRaisesRegex(
            ValueError,
            expected_regex="the number of samples and the number of model coefficients are equal",
        ):
            estimateDispersionsGeneEst(dds)

        dds = makeExampleDESeqDataSet(
            n=100,
            m=4,
            dispMeanRel=lambda x: 0.001 + x / 1e3,
            interceptMean=8,
            interceptSD=2,
        )
        dds = dds.estimateSizeFactors()
        dds.var["dispGeneEst"] = np.repeat(1e-7, 100)
        with self.assertRaisesRegex(
            ValueError,
            expected_regex="all gene-wise dispersion estimates are within 2 orders of magnitude",
        ):
            estimateDispersionsFit(dds)
        dds = estimateDispersionsGeneEst(dds)
        # TODO expect message "note: fitType='parametric', but the dispersion trend was not well captured"
        # estimateDispersionsFit(dds)

        dds = makeExampleDESeqDataSet(n=100, m=4)
        dds = dds.estimateSizeFactors()
        dds.var["dispGeneEst"] = np.repeat(1e-7, 100)
        dds.setDispFunction(lambda x: 1e-6)
        with self.assertWarnsRegex(
            UserWarning,
            expected_regex="all genes have dispersion estimates < 1e-06, returning disp = 1e-07",
        ):
            estimateDispersionsMAP(dds)

        dds = makeExampleDESeqDataSet(n=100, m=4)
        dds = dds.estimateSizeFactors()
        dds.obs["condition"] = Factor(dds.obs["condition"]).add_categories("C")
        dds.design = "~condition"
        with self.assertRaisesRegex(
            ValueError, expected_regex="the model matrix is not full rank"
        ):
            dds.estimateDispersions()
        dds.obs["condition"] = Factor(dds.obs["condition"]).droplevels()
        dds.obs["group"] = dds.obs["condition"]
        dds.design = "~ group + condition"
        with self.assertRaisesRegex(
            ValueError, expected_regex="the model matrix is not full rank"
        ):
            dds.estimateDispersions()

        dds = makeExampleDESeqDataSet(n=100, m=2)
        with self.assertRaisesRegex(
            ValueError,
            expected_regex="The design matrix has the same number of samples and coefficients to fit",
        ):
            DESeq(dds)

    def test_dispersion_fitting(self):
        """test that the fitting of dispersion gives the expected values using various methods"""
        # test the optimization of the logarithm of dispersion (alpha)
        # parameter with Cox-Reid adjustment and prior distribution.
        # also test the derivatives of the log posterior w.r.t. log alpha
        m = 10
        # y = scipy.stats.poisson.rvs(20, size=m, random_state=42)
        y = np.array([17, 25, 25, 21, 13, 22, 23, 22, 18, 26])
        sf = np.ones(m)
        condition = Factor(np.repeat([0, 1], [m / 2, m / 2]))
        x = patsy.dmatrix("1+condition", data={"condition": condition})

        lambda_ = 2
        alpha = 0.5

        # make a DESeqDataSet but don't use the design formula
        # instead we supply directly a model matrix
        dds = DESeqDataSet(
            y[:, None],
            clinicalData=pd.DataFrame({"condition": condition}),
            design="~condition",
        )
        dds.sizeFactors = sf
        dds.var["dispersion"] = alpha
        dds.var["baseMean"] = np.mean(y)

        # for testing we convert beta to the natural log scale:
        # convert lambda from log to log2 scale by multiplying by log(2)**2
        # then convert beta back from log2 to log scale by multiplying by log(2)
        betaDESeq = (
            np.log(2)
            * fitNbinomGLMs(dds, lambda_=[0, lambda_ * np.log(2) ** 2], modelMatrix=x)[
                "betaMatrix"
            ]
        )
        log_alpha_prior_mean = 0.5
        log_alpha_prior_sigmasq = 1
        mu_hat = np.exp(x @ betaDESeq.T).values.squeeze()

        dispRes = fitDisp(
            y[:, None],
            x=x,
            mu_hat=mu_hat[:, None],
            log_alpha=0,
            log_alpha_prior_mean=log_alpha_prior_mean,
            log_alpha_prior_sigmasq=log_alpha_prior_sigmasq,
            min_log_alpha=np.log(1e-8),
            kappa_0=1,
            tol=1e-16,
            maxit=100,
            usePrior=True,
            weights=np.ones((len(y), 1)),
            useWeights=False,
            weightThreshold=1e-2,
            useCR=True,
        )

        # maximum a posteriori (MAP) estimate from DESeq
        dispDESeq = dispRes["log_alpha"]

        # MAP estimate using optim
        def logPost(log_alpha):
            alpha = np.exp(log_alpha)
            w = np.diag(1 / (1 / mu_hat**2 * (mu_hat + alpha * mu_hat**2)))
            logLike = np.sum(dnbinom_mu(y, mu=mu_hat, size=1 / alpha, log=True))
            coxReid = -0.5 * np.linalg.slogdet(x.T @ w @ x)[1]
            logPrior = dnorm(
                log_alpha,
                log_alpha_prior_mean,
                np.sqrt(log_alpha_prior_sigmasq),
                log=True,
            )
            res = logLike + coxReid + logPrior
            return res

        dispOptim = minimize(lambda p: -logPost(p), 0).x

        self.assertTrue(np.allclose(dispDESeq, dispOptim))

        # check derivatives:

        # from Ted Harding https://stat.ethz.ch/pipermail/r-help/2007-September/140013.html
        def num_deriv(f, x, h=0.001):
            return (f(x + h / 2) - f(x - h / 2)) / h

        def num_2nd_deriv(f, x, h=0.001):
            return (f(x + h) - 2 * f(x) + f(x - h)) / h**2

        # first derivative of log posterior wrt log alpha at start
        dispDerivDESeq = dispRes["initial_dlp"]
        dispDerivNum = num_deriv(logPost, 0)
        self.assertTrue(np.allclose(dispDerivDESeq, dispDerivNum))

        # second derivative at finish
        dispD2DESeq = dispRes["last_d2lp"]
        dispD2Num = num_2nd_deriv(logPost, dispRes["log_alpha"])
        self.assertTrue(np.allclose(dispD2DESeq, dispD2Num))

        # test fit alternative
        dds = makeExampleDESeqDataSet()
        dds = dds.estimateSizeFactors()
        # ddsLocal = dds.copy().estimateDispersions(fitType="local")
        ddsMean = dds.copy().estimateDispersions(fitType="mean")
        ddsMed = estimateDispersionsGeneEst(dds.copy())
        useForMedian = ddsMed.var["dispGeneEst"] > 1e-7
        medianDisp = np.nanmedian(ddsMed.var["dispGeneEst"][useForMedian])
        ddsMed.setDispFunction(lambda x: medianDisp)
        ddsMed = estimateDispersionsMAP(ddsMed)

        # test iterative
        dds = makeExampleDESeqDataSet(m=50, n=100, betaSD=1, interceptMean=8, seed=42)
        dds = dds.estimateSizeFactors()
        dds = estimateDispersionsGeneEst(dds, niter=5)
        dds = dds[:, ~dds.var["allZero"]]
        self.assertTrue(
            np.allclose(dds.var["trueDisp"], dds.var["dispGeneEst"], atol=0.7)
        )



# -----------------------------------------------------------------------------
# Copyright (C) 2022-2023 M. Colange

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
# -----------------------------------------------------------------------------

# This file wraps scipy.stats functions into R-like functions.


import numpy as np
from scipy.stats import norm, gamma, nbinom


def rbinom(n, size, prob=None, mu=None, seed=None):
    if prob is None and mu is None:
        raise ValueError("exactly one of prob and mu must be provided")
    if prob is not None and mu is not None:
        raise ValueError("exactly one of prob and mu must be provided")

    if prob is None:
        prob = size / (size + mu)

    return nbinom.rvs(size, prob, size=n, random_state=seed)


def pnbinom(q, size, prob=None, mu=None, lower_tail=True):
    """
    Distribution function of the negative binomial distribution.

    Parameters
    ----------
    q           vector of quantiles
    size        dispersion parameter. Must be strictly positive but not necessarily an integer
    prob        probablity of success in each trial. 0 < prob <= 1
    mu          mean (alternative characterization)
    lower_tail  if True, probabilities are P[X <= x], otherwise P[X > x]
    """
    if prob is None and mu is None:
        raise ValueError("exactly one of prob and mu must be provided")
    if prob is not None and mu is not None:
        raise ValueError("exactly one of prob and mu must be provided")

    if prob is None:
        prob = size / (size + mu)

    if lower_tail:
        return nbinom.cdf(q, size, prob)
    else:
        return nbinom.sf(q, size, prob)


def qnbinom(p, size, prob=None, mu=None, lower_tail=True):
    """
    Quantile function of the negative binomial distribution.

    Parameters
    ----------
    p           vector of probabilities
    size        dispersion parameter. Must be strictly positive but not necessarily an integer
    prob        probablity of success in each trial. 0 < prob <= 1
    mu          mean (alternative characterization)
    lower_tail  if True, probabilities are P[X <= x], otherwise P[X > x]
    """
    if prob is None and mu is None:
        raise ValueError("exactly one of prob and mu must be provided")
    if prob is not None and mu is not None:
        raise ValueError("exactly one of prob and mu must be provided")

    if prob is None:
        prob = size / (size + mu)

    if lower_tail:
        return nbinom.ppf(p, size, prob)
    else:
        return nbinom.isf(p, size, prob)


def qnorm(p, mean=0, sd=1, lower_tail=True, log_p=False):
    """
    Quantile function of the normal distribution.

    Parameters
    ----------
    p           vector of probabilities
    mean        vector of means
    sd          vector of standard deviations
    log_p       if True, probabilities are given in log
    lower_tail  if True, probabilities are P[X <= x], otherwise P[X > x]
    """
    if log_p:
        p = np.exp(p)

    if lower_tail:
        return norm.ppf(p, loc=mean, scale=sd)
    else:
        return norm.isf(p, loc=mean, scale=sd)


def pgamma(q, shape, scale=1, lower_tail=True, log_p=False):
    """
    Distribution function of the gamma distribution.

    Parameters
    ----------
    q           vector of quantiles
    shape       vector of shapes (must be positive)
    scale       vector of scales (must be strictly positive)
    log_p       if True, probabilities are given in log
    lower_tail  if True, probabilities are P[X <= x], otherwise P[X > x]
    """
    if lower_tail and log_p:
        f = gamma.logcdf
    elif lower_tail and not log_p:
        f = gamma.cdf
    elif not lower_tail and log_p:
        f = gamma.logsf
    elif not lower_tail and not log_p:
        f = gamma.sf

    return f(q, shape, scale=scale)


def qgamma(p, shape, scale=1, lower_tail=True, log_p=False):
    """
    Quantile function of the gamma distribution.

    Parameters
    ----------
    p           vector of probabilities
    shape       vector of shapes (must be positive)
    scale       vector of scales (must be strictly positive)
    log_p       if True, probabilities are given in log
    lower_tail  if True, probabilities are P[X <= x], otherwise P[X > x]
    """
    if log_p:
        p = np.exp(p)

    if lower_tail:
        return gamma.ppf(p, shape, scale=scale)
    else:
        return gamma.isf(p, shape, scale=scale)



import unittest
import numpy as np
import pandas as pd
import scipy.stats
from scipy.optimize import minimize

from inmoose.utils import Factor, dnbinom_mu, dnorm
from inmoose.deseq2 import DESeqDataSet
from inmoose.deseq2.fitNbinomGLMs import fitNbinomGLMs


class Test(unittest.TestCase):
    def test_betaFitting(self):
        """test that estimates of beta fit from various methods are equal"""
        m = 10
        y = scipy.stats.poisson.rvs(20, size=m, random_state=42)
        sf = np.ones(m)
        condition = Factor(
            [0 for i in range(int(m / 2))] + [1 for i in range(int(m / 2))]
        )
        x = np.hstack([np.repeat(1, m), np.repeat(0, m / 2), np.repeat(1, m / 2)])
        x = np.ones((m, 2))
        x[: int(m / 2), 1] = 0
        lambda_ = 2
        alpha = 0.5

        dds = DESeqDataSet(
            y[:, None],
            clinicalData=pd.DataFrame({"condition": condition}),
            design="~condition",
        )
        dds.sizeFactors = sf
        dds.var["dispersion"] = alpha
        dds.var["baseMean"] = np.mean(y)

        # for testing we convert beta to the natural log scale:
        # convert lambda_ from log to log2 scale by multiplying by log(2)**2
        # then convert beta back from log2 to log scale by multiplying by log(2)
        betaDESeq = (
            np.log(2)
            * fitNbinomGLMs(dds, lambda_=[0, lambda_ * np.log(2) ** 2])[
                "betaMatrix"
            ].iloc[0, :]
        )

        # the IRLS algorithm
        betaIRLS = np.array([1, 1])
        for t in range(100):
            mu_hat = sf * np.exp(x @ betaIRLS)
            w = np.diag(1 / (1 / mu_hat**2 * (mu_hat + alpha * mu_hat**2)))
            z = np.log(mu_hat / sf) + (y - mu_hat) / mu_hat
            ridge = np.diag([0, lambda_])
            betaIRLS = np.linalg.inv(x.T @ w @ x + ridge) @ x.T @ w @ z

        # using minimize
        def objectiveFn(p):
            mu = np.exp(x @ p)
            logLike = np.sum(dnbinom_mu(y, mu=mu, size=1 / alpha, log=True))
            prior = dnorm(p[1], 0, np.sqrt(1 / lambda_), log=True)
            return -(logLike + prior)

        betaOptim = minimize(objectiveFn, np.array([0.1, 0.1])).x

        self.assertTrue(np.allclose(betaDESeq, betaIRLS))
        self.assertTrue(np.allclose(betaDESeq, betaOptim))



import unittest
import numpy as np
import patsy

from inmoose.utils import Factor
from inmoose.deseq2 import makeExampleDESeqDataSet, DESeq


@unittest.skip("lfcShrink is not implemented")
class Test(unittest.TestCase):
    def test_interation_errors(self):
        """test that interactions throw errors"""
        dds = makeExampleDESeqDataSet(n=100, m=8)
        dds.obs["group"] = Factor(np.repeat(["X", "Y"], dds.n_obs / 2))
        dds.design = "~ condition + group + condition:group"
        dds = DESeq(dds)
        self.assertEqual(dds.resultsNames()[3], "conditionB.groupY")

        # interactions error
        with self.assertRaisesRegex(
            ValueError, expected_regex="designs with interations"
        ):
            DESeq(dds, betaPrior=True)

        # also lfcShrink
        res = dds.results(name="conditionB.groupY")
        with self.assertRaises(NotImplementedError):
            res = lfcShrink(dds, coef=4, res=res, type="normal")

        res = dds.results(contrast=["condition", "B", "A"])
        with self.assertRaises(NotImplementedError):
            res = lfcShrink(
                dds, contrast=["condition", "B", "A"], res=res, type="normal"
            )

        # however, this is allowed
        dds2 = dds.copy()
        dds2.design = patsy.dmatrix("~ condition + group + condition:group", dds2.obs)
        dds2 = DESeq(dds2)
        res2 = dds2.results(name="conditionB.groupY")
        res2 = lfcShrink(dds2, coef=4, res=res, type="normal")



# -----------------------------------------------------------------------------
# Copyright (C) 2008-2022 Yunshun Chen, Aaron TL Lun, Davis J McCarthy, Matthew E Ritchie, Belinda Phipson, Yifang Hu, Xiaobei Zhou, Mark D Robinson, Gordon K Smyth
# Copyright (C) 2022-2023 Maximilien Colange

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
# -----------------------------------------------------------------------------

# This file is based on the file 'R/validDGEList.R' of the Bioconductor edgeR package (version 3.38.4).


import numpy as np
from ..utils import gl


def validDGEList(y):
    """
    Check for standard components of DGEList object

    NB: Modifies :code:`y` in place.

    Arguments
    ---------
    y : DGEList
        the object to check for validity

    Returns
    -------
    DGEList
        the input :code:`y`, with missing components added
    """
    if y.counts is None:
        raise RuntimeError("No count matrix")
    y.counts = np.asarray(y.counts, order="F")
    nlib = y.counts.shape[1]
    if (y.samples.group.values == None).any():
        y.samples.group = gl(1, nlib)
    if (y.samples.lib_size.values == None).any():
        y.samples.lib_size = y.counts.sum(axis=0)
    if (y.samples.norm_factors.values == None).any():
        y.samples.norm_factors = np.ones(nlib)

    return y



# -----------------------------------------------------------------------------
# Copyright (C) 2008-2022 Yunshun Chen, Aaron TL Lun, Davis J McCarthy, Matthew E Ritchie, Belinda Phipson, Yifang Hu, Xiaobei Zhou, Mark D Robinson, Gordon K Smyth
# Copyright (C) 2022-2023 Maximilien Colange

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
# -----------------------------------------------------------------------------

# This file is based on the file 'R/DGEList.R' of the Bioconductor edgeR package (version 3.38.4).

import logging
import numpy as np
import pandas as pd

from .edgepy_cpp import is_integer_array
from ..utils import Factor
from .utils import _isAllZero


class DGEList(object):
    """
    A class for storing read counts and associated information from digital
    gene expression or sequencing technologies.

    Attributes
    ----------
    counts : ndarray
        matrix of read counts, one row per gene and one column per sample
    samples : DataFrame
        dataframe with a row for each sample and columns :code:`group`,
        :code:`lib_size` and :code:`norm_factors` containing the group labels,
        library sizes and normalization factors.
        Other columns can be optionally added to give more detailed sample
        information.
    common_dispersion : float, optional
        the overall dispersion estimate
    tagwise_dispersion : ndarray, optional
        genewise dispersion estimates for each gene ("tag" and "gene" are
        synonymous here)
    trended_dispersion : ndarray, optional
        trended dispersion estimates for each gene
    offset : array_like, optional
        matrix of same shape as :code:`counts` giving offsets for log-linear
        models
    genes : DataFrame, optional
        annotation information for each gene. Same number of rows as
        :code:`counts`
    AveLogCPM : ndarray, optional
        average log2 counts per million for each gene
    """

    from .aveLogCPM import aveLogCPM_DGEList as aveLogCPM
    from .estimateGLMCommonDisp import (
        estimateGLMCommonDisp_DGEList as estimateGLMCommonDisp,
    )
    from .estimateGLMTagwiseDisp import (
        estimateGLMTagwiseDisp_DGEList as estimateGLMTagwiseDisp,
    )
    from .glmFit import glmFit_DGEList as glmFit
    from .predFC import predFC_DGEList as predFC
    from .splitIntoGroups import splitIntoGroups_DGEList as splitIntoGroups

    def __init__(
        self,
        counts,
        lib_size=None,
        norm_factors=None,
        samples=None,
        group=None,
        genes=None,
        remove_zeroes=False,
    ):
        """
        Construct DGEList object from components with some checking

        Arguments
        ---------
        counts : array_like
            matrix of counts
        lib_size : array_like, optional
            vector of total counts (sequence depth) for each library
        norm_factors : array_like, optional
            vector of normalization factors that modify the library sizes
        samples : DataFrame, optional
            information for each sample
        group : array_like or Factor, optional
            vector or factor giving the experimental group/condition for each
            sample/library
        genes : DataFrame, optional
            annotation information for each gene
        remove_zeroes : bool
            whether to remove rows that have 0 total count
        """

        # Check counts
        counts = np.asarray(counts, order="F")
        if len(counts.shape) != 2:
            raise ValueError("'counts' is not a matrix!")
        try:
            is_integer_array(counts)
        except RuntimeError:
            raise ValueError("non-numeric values found in 'counts'")

        (ntags, nlibs) = counts.shape
        # TODO fill in colnames
        # TODO fill in rownames
        _isAllZero(
            counts
        )  # don't really care about all-zeroes, but do want to protect against NaN, infinite and negative values

        # Check lib_size
        if lib_size is None:
            lib_size = counts.sum(axis=0)
        lib_size = np.asarray(lib_size, order="F")
        # TODO check that lib_size is numeric
        if nlibs != len(lib_size):
            raise ValueError(
                "length of 'lib_size' must be equal to the number of columns in 'counts'"
            )
        minlibsize = np.min(lib_size)
        # TODO check lib_size for NaN
        if minlibsize < 0:
            raise ValueError("negative library size not permitted")
        if minlibsize == 0:
            if np.logical_and(lib_size == 0, (counts.sum(axis=0) > 0)).any():
                raise ValueError("library size set to zero but counts are nonzero")
            logging.warnings.warn("library size of zero detected")

        # Check norm_factors
        if norm_factors is None:
            norm_factors = np.ones(nlibs)
        # TODO check that norm_factors is numeric
        if nlibs != len(norm_factors):
            raise ValueError(
                "Length of 'norm_factors' must be equal to the number of columns in 'counts'"
            )
        minnf = norm_factors.min()
        # TODO check norm_factors for NaN
        if minnf <= 0:
            raise ValueError("norm factors should be positive")
        # TODO check that norm factors multiply to 1 (more or less 1-e6)

        # Check samples
        if samples is not None:
            if samples.shape()[0] != nlibs:
                raise ValueError(
                    "Number of rows in 'samples' must be equal to the number of columns in 'counts'"
                )

        # Get group from samples if appropriate
        if (
            group is None
            and samples is not None
            and (samples.group.values != None).all()
        ):
            group = samples.group.values
            samples = samples.drop(columns=["group"])

        # Check group
        if group is None:
            group = np.ones(nlibs)
            group = Factor(group)
        if len(group) != nlibs:
            raise ValueError(
                "Length of 'group' must be equal to the number of columns in 'counts'"
            )

        # Make data frame of sample informations
        # in R, acts as a dictionnary of info for each row. R allows duplicated column names (with a warning message)
        sam = pd.DataFrame(
            dict(group=group, lib_size=lib_size, norm_factors=norm_factors)
        )
        if samples is not None:
            sam = pd.concat([sam, samples], axis=1)
        samples = sam
        # TODO set row names in 'samples' according to column names in 'counts'

        # make object
        self.counts = counts
        self.samples = samples
        self.design = None
        self.weights = None
        self.common_dispersion = None
        self.tagwise_dispersion = None
        self.trended_dispersion = None
        self.offset = None
        self.genes = None
        self.prior_df = None
        self.AveLogCPM = None

        # TODO Add data frame of gene information (should set self.genes)
        if genes is not None:
            if len(genes) != ntags:
                raise ValueError("Counts and genes have different number of rows")
            # TODO check for duplicated gene names (?)
            self.genes = genes

        # TODO Remove rows with all zeroes
        if remove_zeroes:
            raise NotImplementedError

    def getOffset(self):
        """
        Extract offset vector or matrix from data object and optional arguments.
        By default, offset is constructed from the lib_size and norm_factors
        """
        if self.offset is not None:
            return self.offset

        lib_size = self.samples.lib_size
        norm_factors = self.samples.norm_factors
        if (norm_factors.values != None).all():
            lib_size = lib_size * norm_factors
        return np.log(lib_size)

    def getDispersion(self):
        """
        Get most complex dispersion values from DGEList object
        """
        if self.tagwise_dispersion is not None:
            dispersion = self.tagwise_dispersion
        elif self.trended_dispersion is not None:
            dispersion = self.trended_dispersion
        elif self.common_dispersion is not None:
            dispersion = self.common_dispersion
        else:
            dispersion = None

        return dispersion
